#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pickle
import sys

from taskflow import exceptions
from taskflow import test
from taskflow.tests import utils as test_utils
from taskflow.types import failure


def _captured_failure(msg):
    try:
        raise RuntimeError(msg)
    except Exception:
        return failure.Failure()


def _make_exc_info(msg):
    try:
        raise RuntimeError(msg)
    except Exception:
        return sys.exc_info()


class GeneralFailureObjTestsMixin:

    def test_captures_message(self):
        self.assertEqual('Woot!', self.fail_obj.exception_str)

    def test_str(self):
        self.assertEqual('Failure: RuntimeError: Woot!',
                         str(self.fail_obj))

    def test_exception_types(self):
        self.assertEqual(test_utils.RUNTIME_ERROR_CLASSES[:-2],
                         list(self.fail_obj))

    def test_pformat_no_traceback(self):
        text = self.fail_obj.pformat()
        self.assertNotIn("Traceback", text)

    def test_check_str(self):
        val = 'Exception'
        self.assertEqual(val, self.fail_obj.check(val))

    def test_check_str_not_there(self):
        val = 'ValueError'
        self.assertIsNone(self.fail_obj.check(val))

    def test_check_type(self):
        self.assertIs(self.fail_obj.check(RuntimeError), RuntimeError)

    def test_check_type_not_there(self):
        self.assertIsNone(self.fail_obj.check(ValueError))


class CaptureFailureTestCase(test.TestCase, GeneralFailureObjTestsMixin):

    def setUp(self):
        super().setUp()
        self.fail_obj = _captured_failure('Woot!')

    def test_captures_value(self):
        self.assertIsInstance(self.fail_obj.exception, RuntimeError)

    def test_captures_exc_info(self):
        exc_info = self.fail_obj.exc_info
        self.assertEqual(3, len(exc_info))
        self.assertEqual(RuntimeError, exc_info[0])
        self.assertIs(exc_info[1], self.fail_obj.exception)

    def test_reraises(self):
        self.assertRaisesRegex(RuntimeError, '^Woot!$', self.fail_obj.reraise)


class ReCreatedFailureTestCase(test.TestCase, GeneralFailureObjTestsMixin):

    def setUp(self):
        super().setUp()
        fail_obj = _captured_failure('Woot!')
        self.fail_obj = failure.Failure(exception_str=fail_obj.exception_str,
                                        traceback_str=fail_obj.traceback_str,
                                        exc_type_names=list(fail_obj))

    def test_value_lost(self):
        self.assertIsNone(self.fail_obj.exception)

    def test_no_exc_info(self):
        self.assertIsNone(self.fail_obj.exc_info)

    def test_pformat_traceback(self):
        text = self.fail_obj.pformat(traceback=True)
        self.assertIn("Traceback (most recent call last):", text)

    def test_reraises(self):
        exc = self.assertRaises(exceptions.WrappedFailure,
                                self.fail_obj.reraise)
        self.assertIs(exc.check(RuntimeError), RuntimeError)

    def test_no_type_names(self):
        fail_obj = _captured_failure('Woot!')
        fail_obj = failure.Failure(exception_str=fail_obj.exception_str,
                                   traceback_str=fail_obj.traceback_str,
                                   exc_type_names=[])
        self.assertEqual([], list(fail_obj))
        self.assertEqual("Failure: Woot!", fail_obj.pformat())


class FromExceptionTestCase(test.TestCase, GeneralFailureObjTestsMixin):

    def setUp(self):
        super().setUp()
        self.fail_obj = failure.Failure.from_exception(RuntimeError('Woot!'))

    def test_pformat_no_traceback(self):
        text = self.fail_obj.pformat(traceback=True)
        self.assertIn("Traceback not available", text)


class FailureObjectTestCase(test.TestCase):

    def test_invalids(self):
        f = {
            'exception_str': 'blah',
            'traceback_str': 'blah',
            'exc_type_names': [],
        }
        self.assertRaises(exceptions.InvalidFormat,
                          failure.Failure.validate, f)
        f = {
            'exception_str': 'blah',
            'exc_type_names': ['Exception'],
        }
        self.assertRaises(exceptions.InvalidFormat,
                          failure.Failure.validate, f)
        f = {
            'exception_str': 'blah',
            'traceback_str': 'blah',
            'exc_type_names': ['Exception'],
            'version': -1,
        }
        self.assertRaises(exceptions.InvalidFormat,
                          failure.Failure.validate, f)

    def test_valid_from_dict_to_dict(self):
        f = _captured_failure('Woot!')
        d_f = f.to_dict()
        failure.Failure.validate(d_f)
        f2 = failure.Failure.from_dict(d_f)
        self.assertTrue(f.matches(f2))

    def test_bad_root_exception(self):
        f = _captured_failure('Woot!')
        d_f = f.to_dict()
        d_f['exc_type_names'] = ['Junk']
        self.assertRaises(exceptions.InvalidFormat,
                          failure.Failure.validate, d_f)

    def test_valid_from_dict_to_dict_2(self):
        f = _captured_failure('Woot!')
        d_f = f.to_dict()
        d_f['exc_type_names'] = ['RuntimeError', 'Exception', 'BaseException']
        failure.Failure.validate(d_f)

    def test_cause_exception_args(self):
        f = _captured_failure('Woot!')
        d_f = f.to_dict()
        self.assertEqual(1, len(d_f['exc_args']))
        self.assertEqual(("Woot!",), d_f['exc_args'])
        f2 = failure.Failure.from_dict(d_f)
        self.assertEqual(f.exception_args, f2.exception_args)

    def test_dont_catch_base_exception(self):
        try:
            raise SystemExit()
        except BaseException:
            self.assertRaises(TypeError, failure.Failure)

    def test_unknown_argument(self):
        exc = self.assertRaises(TypeError, failure.Failure,
                                exception_str='Woot!',
                                traceback_str=None,
                                exc_type_names=['Exception'],
                                hi='hi there')
        expected = "Failure.__init__ got unexpected keyword argument(s): hi"
        self.assertEqual(expected, str(exc))

    def test_empty_does_not_reraise(self):
        self.assertIsNone(failure.Failure.reraise_if_any([]))

    def test_reraises_one(self):
        fls = [_captured_failure('Woot!')]
        self.assertRaisesRegex(RuntimeError, '^Woot!$',
                               failure.Failure.reraise_if_any, fls)

    def test_reraises_several(self):
        fls = [
            _captured_failure('Woot!'),
            _captured_failure('Oh, not again!')
        ]
        exc = self.assertRaises(exceptions.WrappedFailure,
                                failure.Failure.reraise_if_any, fls)
        self.assertEqual(fls, list(exc))

    def test_failure_copy(self):
        fail_obj = _captured_failure('Woot!')

        copied = fail_obj.copy()
        self.assertIsNot(fail_obj, copied)
        self.assertEqual(fail_obj, copied)
        self.assertTrue(fail_obj.matches(copied))

    def test_failure_copy_recaptured(self):
        captured = _captured_failure('Woot!')
        fail_obj = failure.Failure(exception_str=captured.exception_str,
                                   traceback_str=captured.traceback_str,
                                   exc_type_names=list(captured))
        copied = fail_obj.copy()
        self.assertIsNot(fail_obj, copied)
        self.assertEqual(fail_obj, copied)
        self.assertFalse(fail_obj != copied)
        self.assertTrue(fail_obj.matches(copied))

    def test_recaptured_not_eq(self):
        captured = _captured_failure('Woot!')
        fail_obj = failure.Failure(exception_str=captured.exception_str,
                                   traceback_str=captured.traceback_str,
                                   exc_type_names=list(captured),
                                   exc_args=list(captured.exception_args))
        self.assertFalse(fail_obj == captured)
        self.assertTrue(fail_obj != captured)
        self.assertTrue(fail_obj.matches(captured))

    def test_two_captured_eq(self):
        captured = _captured_failure('Woot!')
        captured2 = _captured_failure('Woot!')
        self.assertEqual(captured, captured2)

    def test_two_recaptured_neq(self):
        captured = _captured_failure('Woot!')
        fail_obj = failure.Failure(exception_str=captured.exception_str,
                                   traceback_str=captured.traceback_str,
                                   exc_type_names=list(captured))
        new_exc_str = captured.exception_str.replace('Woot', 'w00t')
        fail_obj2 = failure.Failure(exception_str=new_exc_str,
                                    traceback_str=captured.traceback_str,
                                    exc_type_names=list(captured))
        self.assertNotEqual(fail_obj, fail_obj2)
        self.assertFalse(fail_obj2.matches(fail_obj))

    def test_compares_to_none(self):
        captured = _captured_failure('Woot!')
        self.assertIsNotNone(captured)
        self.assertFalse(captured.matches(None))

    def test_pformat_traceback(self):
        captured = _captured_failure('Woot!')
        text = captured.pformat(traceback=True)
        self.assertIn("Traceback (most recent call last):", text)

    def test_pformat_traceback_captured_no_exc_info(self):
        captured = _captured_failure('Woot!')
        captured = failure.Failure.from_dict(captured.to_dict())
        text = captured.pformat(traceback=True)
        self.assertIn("Traceback (most recent call last):", text)

    def test_no_capture_exc_args(self):
        captured = _captured_failure(Exception("I am not valid JSON"))
        fail_obj = failure.Failure(exception_str=captured.exception_str,
                                   traceback_str=captured.traceback_str,
                                   exc_type_names=list(captured),
                                   exc_args=list(captured.exception_args))
        fail_json = fail_obj.to_dict(include_args=False)
        self.assertNotEqual(fail_obj.exception_args, fail_json['exc_args'])
        self.assertEqual(fail_json['exc_args'], tuple())


class WrappedFailureTestCase(test.TestCase):

    def test_simple_iter(self):
        fail_obj = _captured_failure('Woot!')
        wf = exceptions.WrappedFailure([fail_obj])
        self.assertEqual(1, len(wf))
        self.assertEqual([fail_obj], list(wf))

    def test_simple_check(self):
        fail_obj = _captured_failure('Woot!')
        wf = exceptions.WrappedFailure([fail_obj])
        self.assertEqual(RuntimeError, wf.check(RuntimeError))
        self.assertIsNone(wf.check(ValueError))

    def test_two_failures(self):
        fls = [
            _captured_failure('Woot!'),
            _captured_failure('Oh, not again!')
        ]
        wf = exceptions.WrappedFailure(fls)
        self.assertEqual(2, len(wf))
        self.assertEqual(fls, list(wf))

    def test_flattening(self):
        f1 = _captured_failure('Wrap me')
        f2 = _captured_failure('Wrap me, too')
        f3 = _captured_failure('Woot!')
        try:
            raise exceptions.WrappedFailure([f1, f2])
        except Exception:
            fail_obj = failure.Failure()

        wf = exceptions.WrappedFailure([fail_obj, f3])
        self.assertEqual([f1, f2, f3], list(wf))


class NonAsciiExceptionsTestCase(test.TestCase):

    def test_exception_with_non_ascii_str(self):
        bad_string = chr(200)
        excp = ValueError(bad_string)
        fail = failure.Failure.from_exception(excp)
        self.assertEqual(str(excp), fail.exception_str)
        expected = 'Failure: ValueError: \xc8'
        self.assertEqual(expected, str(fail))

    def test_exception_non_ascii_unicode(self):
        hi_ru = 'привет'
        fail = failure.Failure.from_exception(ValueError(hi_ru))
        self.assertEqual(hi_ru, fail.exception_str)
        self.assertIsInstance(fail.exception_str, str)
        self.assertEqual('Failure: ValueError: %s' % hi_ru,
                         str(fail))

    def test_wrapped_failure_non_ascii_unicode(self):
        hi_cn = '嗨'
        fail = ValueError(hi_cn)
        self.assertEqual(hi_cn, str(fail))
        fail = failure.Failure.from_exception(fail)
        wrapped_fail = exceptions.WrappedFailure([fail])
        expected_result = ("WrappedFailure: "
                           "[Failure: ValueError: %s]" % (hi_cn))
        self.assertEqual(expected_result, str(wrapped_fail))

    def test_failure_equality_with_non_ascii_str(self):
        bad_string = chr(200)
        fail = failure.Failure.from_exception(ValueError(bad_string))
        copied = fail.copy()
        self.assertEqual(fail, copied)

    def test_failure_equality_non_ascii_unicode(self):
        hi_ru = 'привет'
        fail = failure.Failure.from_exception(ValueError(hi_ru))
        copied = fail.copy()
        self.assertEqual(fail, copied)


class FailureCausesTest(test.TestCase):

    @classmethod
    def _raise_many(cls, messages):
        if not messages:
            return
        msg = messages.pop(0)
        e = RuntimeError(msg)
        try:
            cls._raise_many(messages)
            raise e
        except RuntimeError as e1:
            raise e from e1

    def test_causes(self):
        f = None
        try:
            self._raise_many(["Still still not working",
                              "Still not working", "Not working"])
        except RuntimeError:
            f = failure.Failure()

        self.assertIsNotNone(f)
        self.assertEqual(2, len(f.causes))
        self.assertEqual("Still not working", f.causes[0].exception_str)
        self.assertEqual("Not working", f.causes[1].exception_str)

        f = f.causes[0]
        self.assertEqual(1, len(f.causes))
        self.assertEqual("Not working", f.causes[0].exception_str)

        f = f.causes[0]
        self.assertEqual(0, len(f.causes))

    def test_causes_to_from_dict(self):
        f = None
        try:
            self._raise_many(["Still still not working",
                              "Still not working", "Not working"])
        except RuntimeError:
            f = failure.Failure()

        self.assertIsNotNone(f)
        d_f = f.to_dict()
        failure.Failure.validate(d_f)
        f = failure.Failure.from_dict(d_f)
        self.assertEqual(2, len(f.causes))
        self.assertEqual("Still not working", f.causes[0].exception_str)
        self.assertEqual("Not working", f.causes[1].exception_str)

        f = f.causes[0]
        self.assertEqual(1, len(f.causes))
        self.assertEqual("Not working", f.causes[0].exception_str)

        f = f.causes[0]
        self.assertEqual(0, len(f.causes))

    def test_causes_pickle(self):
        f = None
        try:
            self._raise_many(["Still still not working",
                              "Still not working", "Not working"])
        except RuntimeError:
            f = failure.Failure()

        self.assertIsNotNone(f)
        p_f = pickle.dumps(f)
        f = pickle.loads(p_f)

        self.assertEqual(2, len(f.causes))
        self.assertEqual("Still not working", f.causes[0].exception_str)
        self.assertEqual("Not working", f.causes[1].exception_str)

        f = f.causes[0]
        self.assertEqual(1, len(f.causes))
        self.assertEqual("Not working", f.causes[0].exception_str)

        f = f.causes[0]
        self.assertEqual(0, len(f.causes))

    def test_causes_suppress_context(self):
        f = None
        try:
            try:
                self._raise_many(["Still still not working",
                                  "Still not working", "Not working"])
            except RuntimeError as e:
                raise e from None
        except RuntimeError:
            f = failure.Failure()

        self.assertIsNotNone(f)
        self.assertEqual([], list(f.causes))


class ExcInfoUtilsTest(test.TestCase):
    def test_copy_none(self):
        result = failure._copy_exc_info(None)
        self.assertIsNone(result)

    def test_copy_exc_info(self):
        exc_info = _make_exc_info("Woot!")
        result = failure._copy_exc_info(exc_info)
        self.assertIsNot(result, exc_info)
        self.assertIs(result[0], RuntimeError)
        self.assertIsNot(result[1], exc_info[1])
        self.assertIs(result[2], exc_info[2])

    def test_none_equals(self):
        self.assertTrue(failure._are_equal_exc_info_tuples(None, None))

    def test_none_ne_tuple(self):
        exc_info = _make_exc_info("Woot!")
        self.assertFalse(failure._are_equal_exc_info_tuples(None, exc_info))

    def test_tuple_nen_none(self):
        exc_info = _make_exc_info("Woot!")
        self.assertFalse(failure._are_equal_exc_info_tuples(exc_info, None))

    def test_tuple_equals_itself(self):
        exc_info = _make_exc_info("Woot!")
        self.assertTrue(failure._are_equal_exc_info_tuples(exc_info, exc_info))

    def test_typle_equals_copy(self):
        exc_info = _make_exc_info("Woot!")
        copied = failure._copy_exc_info(exc_info)
        self.assertTrue(failure._are_equal_exc_info_tuples(exc_info, copied))



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from stevedore import driver

from taskflow import exceptions as exc
from taskflow import logging
from taskflow.utils import misc


# NOTE(harlowja): this is the entrypoint namespace, not the module namespace.
BACKEND_NAMESPACE = 'taskflow.jobboards'

LOG = logging.getLogger(__name__)


def fetch(name, conf, namespace=BACKEND_NAMESPACE, **kwargs):
    """Fetch a jobboard backend with the given configuration.

    This fetch method will look for the entrypoint name in the entrypoint
    namespace, and then attempt to instantiate that entrypoint using the
    provided name, configuration and any board specific kwargs.

    NOTE(harlowja): to aid in making it easy to specify configuration and
    options to a board the configuration (which is typical just a dictionary)
    can also be a URI string that identifies the entrypoint name and any
    configuration specific to that board.

    For example, given the following configuration URI::

        zookeeper://<not-used>/?a=b&c=d

    This will look for the entrypoint named 'zookeeper' and will provide
    a configuration object composed of the URI's components, in this case that
    is ``{'a': 'b', 'c': 'd'}`` to the constructor of that board
    instance (also including the name specified).
    """
    board, conf = misc.extract_driver_and_conf(conf, 'board')
    LOG.debug('Looking for %r jobboard driver in %r', board, namespace)
    try:
        mgr = driver.DriverManager(namespace, board,
                                   invoke_on_load=True,
                                   invoke_args=(name, conf),
                                   invoke_kwds=kwargs)
        return mgr.driver
    except RuntimeError as e:
        raise exc.NotFound("Could not find jobboard %s" % (board), e)


@contextlib.contextmanager
def backend(name, conf, namespace=BACKEND_NAMESPACE, **kwargs):
    """Fetches a jobboard, connects to it and closes it on completion.

    This allows a board instance to fetched, connected to, and then used in a
    context manager statement with the board being closed upon context
    manager exit.
    """
    jb = fetch(name, conf, namespace=namespace, **kwargs)
    jb.connect()
    with contextlib.closing(jb):
        yield jb



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from automaton import exceptions as excp
from automaton import runners

from taskflow.engines.action_engine import builder
from taskflow.engines.action_engine import compiler
from taskflow.engines.action_engine import executor
from taskflow.engines.action_engine import runtime
from taskflow.patterns import linear_flow as lf
from taskflow import states as st
from taskflow import storage
from taskflow import test
from taskflow.tests import utils as test_utils
from taskflow.types import notifier
from taskflow.utils import persistence_utils as pu


class BuildersTest(test.TestCase):

    def _make_runtime(self, flow, initial_state=None):
        compilation = compiler.PatternCompiler(flow).compile()
        flow_detail = pu.create_flow_detail(flow)
        store = storage.Storage(flow_detail)
        nodes_iter = compilation.execution_graph.nodes(data=True)
        for node, node_attrs in nodes_iter:
            if node_attrs['kind'] in ('task', 'retry'):
                store.ensure_atom(node)
        if initial_state:
            store.set_flow_state(initial_state)
        atom_notifier = notifier.Notifier()
        task_executor = executor.SerialTaskExecutor()
        retry_executor = executor.SerialRetryExecutor()
        task_executor.start()
        self.addCleanup(task_executor.stop)
        r = runtime.Runtime(compilation, store,
                            atom_notifier, task_executor,
                            retry_executor)
        r.compile()
        return r

    def _make_machine(self, flow, initial_state=None):
        runtime = self._make_runtime(flow, initial_state=initial_state)
        machine, memory = runtime.builder.build({})
        machine_runner = runners.FiniteRunner(machine)
        return (runtime, machine, memory, machine_runner)

    def test_run_iterations(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            1, task_cls=test_utils.TaskNoRequiresNoReturns)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        it = machine_runner.run_iter(builder.START)
        prior_state, new_state = next(it)
        self.assertEqual(st.RESUMING, new_state)
        self.assertEqual(0, len(memory.failures))

        prior_state, new_state = next(it)
        self.assertEqual(st.SCHEDULING, new_state)
        self.assertEqual(0, len(memory.failures))

        prior_state, new_state = next(it)
        self.assertEqual(st.WAITING, new_state)
        self.assertEqual(0, len(memory.failures))

        prior_state, new_state = next(it)
        self.assertEqual(st.ANALYZING, new_state)
        self.assertEqual(0, len(memory.failures))

        prior_state, new_state = next(it)
        self.assertEqual(builder.GAME_OVER, new_state)
        self.assertEqual(0, len(memory.failures))
        prior_state, new_state = next(it)
        self.assertEqual(st.SUCCESS, new_state)
        self.assertEqual(0, len(memory.failures))

        self.assertRaises(StopIteration, next, it)

    def test_run_iterations_reverted(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            1, task_cls=test_utils.TaskWithFailure)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = list(machine_runner.run_iter(builder.START))
        prior_state, new_state = transitions[-1]
        self.assertEqual(st.REVERTED, new_state)
        self.assertEqual([], memory.failures)
        self.assertEqual(st.REVERTED,
                         runtime.storage.get_atom_state(tasks[0].name))

    def test_run_iterations_failure(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            1, task_cls=test_utils.NastyFailingTask)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = list(machine_runner.run_iter(builder.START))
        prior_state, new_state = transitions[-1]
        self.assertEqual(st.FAILURE, new_state)
        self.assertEqual(1, len(memory.failures))
        failure = memory.failures[0]
        self.assertTrue(failure.check(RuntimeError))
        self.assertEqual(st.REVERT_FAILURE,
                         runtime.storage.get_atom_state(tasks[0].name))

    def test_run_iterations_suspended(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            2, task_cls=test_utils.TaskNoRequiresNoReturns)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = []
        for prior_state, new_state in machine_runner.run_iter(builder.START):
            transitions.append((new_state, memory.failures))
            if new_state == st.ANALYZING:
                runtime.storage.set_flow_state(st.SUSPENDED)
        state, failures = transitions[-1]
        self.assertEqual(st.SUSPENDED, state)
        self.assertEqual([], failures)

        self.assertEqual(st.SUCCESS,
                         runtime.storage.get_atom_state(tasks[0].name))
        self.assertEqual(st.PENDING,
                         runtime.storage.get_atom_state(tasks[1].name))

    def test_run_iterations_suspended_failure(self):
        flow = lf.Flow("root")
        sad_tasks = test_utils.make_many(
            1, task_cls=test_utils.NastyFailingTask)
        flow.add(*sad_tasks)
        happy_tasks = test_utils.make_many(
            1, task_cls=test_utils.TaskNoRequiresNoReturns, offset=1)
        flow.add(*happy_tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = []
        for prior_state, new_state in machine_runner.run_iter(builder.START):
            transitions.append((new_state, memory.failures))
            if new_state == st.ANALYZING:
                runtime.storage.set_flow_state(st.SUSPENDED)
        state, failures = transitions[-1]
        self.assertEqual(st.SUSPENDED, state)
        self.assertEqual([], failures)

        self.assertEqual(st.PENDING,
                         runtime.storage.get_atom_state(happy_tasks[0].name))
        self.assertEqual(st.FAILURE,
                         runtime.storage.get_atom_state(sad_tasks[0].name))

    def test_builder_manual_process(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            1, task_cls=test_utils.TaskNoRequiresNoReturns)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)
        self.assertRaises(excp.NotInitialized, machine.process_event, 'poke')

        # Should now be pending...
        self.assertEqual(st.PENDING,
                         runtime.storage.get_atom_state(tasks[0].name))

        machine.initialize()
        self.assertEqual(builder.UNDEFINED, machine.current_state)
        self.assertFalse(machine.terminated)
        self.assertRaises(excp.NotFound, machine.process_event, 'poke')
        last_state = machine.current_state

        reaction, terminal = machine.process_event(builder.START)
        self.assertFalse(terminal)
        self.assertIsNotNone(reaction)
        self.assertEqual(st.RESUMING, machine.current_state)
        self.assertRaises(excp.NotFound, machine.process_event, 'poke')

        last_state = machine.current_state
        cb, args, kwargs = reaction
        next_event = cb(last_state, machine.current_state,
                        builder.START, *args, **kwargs)
        reaction, terminal = machine.process_event(next_event)
        self.assertFalse(terminal)
        self.assertIsNotNone(reaction)
        self.assertEqual(st.SCHEDULING, machine.current_state)
        self.assertRaises(excp.NotFound, machine.process_event, 'poke')

        last_state = machine.current_state
        cb, args, kwargs = reaction
        next_event = cb(last_state, machine.current_state,
                        next_event, *args, **kwargs)
        reaction, terminal = machine.process_event(next_event)
        self.assertFalse(terminal)
        self.assertEqual(st.WAITING, machine.current_state)
        self.assertRaises(excp.NotFound, machine.process_event, 'poke')

        # Should now be running...
        self.assertEqual(st.RUNNING,
                         runtime.storage.get_atom_state(tasks[0].name))

        last_state = machine.current_state
        cb, args, kwargs = reaction
        next_event = cb(last_state, machine.current_state,
                        next_event, *args, **kwargs)
        reaction, terminal = machine.process_event(next_event)
        self.assertFalse(terminal)
        self.assertIsNotNone(reaction)
        self.assertEqual(st.ANALYZING, machine.current_state)
        self.assertRaises(excp.NotFound, machine.process_event, 'poke')

        last_state = machine.current_state
        cb, args, kwargs = reaction
        next_event = cb(last_state, machine.current_state,
                        next_event, *args, **kwargs)
        reaction, terminal = machine.process_event(next_event)
        self.assertFalse(terminal)
        self.assertEqual(builder.GAME_OVER, machine.current_state)

        # Should now be done...
        self.assertEqual(st.SUCCESS,
                         runtime.storage.get_atom_state(tasks[0].name))

    def test_builder_automatic_process(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            1, task_cls=test_utils.TaskNoRequiresNoReturns)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = list(machine_runner.run_iter(builder.START))
        self.assertEqual((builder.UNDEFINED, st.RESUMING), transitions[0])
        self.assertEqual((builder.GAME_OVER, st.SUCCESS), transitions[-1])
        self.assertEqual(st.SUCCESS,
                         runtime.storage.get_atom_state(tasks[0].name))

    def test_builder_automatic_process_failure(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(1, task_cls=test_utils.NastyFailingTask)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = list(machine_runner.run_iter(builder.START))
        self.assertEqual((builder.GAME_OVER, st.FAILURE), transitions[-1])
        self.assertEqual(1, len(memory.failures))

    def test_builder_automatic_process_reverted(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(1, task_cls=test_utils.TaskWithFailure)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)

        transitions = list(machine_runner.run_iter(builder.START))
        self.assertEqual((builder.GAME_OVER, st.REVERTED), transitions[-1])
        self.assertEqual(st.REVERTED,
                         runtime.storage.get_atom_state(tasks[0].name))

    def test_builder_expected_transition_occurrences(self):
        flow = lf.Flow("root")
        tasks = test_utils.make_many(
            10, task_cls=test_utils.TaskNoRequiresNoReturns)
        flow.add(*tasks)

        runtime, machine, memory, machine_runner = self._make_machine(
            flow, initial_state=st.RUNNING)
        transitions = list(machine_runner.run_iter(builder.START))

        occurrences = {t: transitions.count(t) for t in transitions}
        self.assertEqual(10, occurrences.get((st.SCHEDULING, st.WAITING)))
        self.assertEqual(10, occurrences.get((st.WAITING, st.ANALYZING)))
        self.assertEqual(9, occurrences.get((st.ANALYZING, st.SCHEDULING)))
        self.assertEqual(1, occurrences.get((builder.GAME_OVER, st.SUCCESS)))
        self.assertEqual(1, occurrences.get((builder.UNDEFINED, st.RESUMING)))

        self.assertEqual(0, len(memory.next_up))
        self.assertEqual(0, len(memory.not_done))
        self.assertEqual(0, len(memory.failures))



#    Copyright (C) 2015 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum
import functools

import redis
from redis import exceptions as redis_exceptions


def _raise_on_closed(meth):

    @functools.wraps(meth)
    def wrapper(self, *args, **kwargs):
        if self.closed:
            raise redis_exceptions.ConnectionError("Connection has been"
                                                   " closed")
        return meth(self, *args, **kwargs)

    return wrapper


class RedisClient(redis.Redis):
    """A redis client that can be closed (and raises on-usage after closed).

    TODO(harlowja): if https://github.com/andymccurdy/redis-py/issues/613 ever
    gets resolved or merged or other then we can likely remove this.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.closed = False

    def close(self):
        self.closed = True
        self.connection_pool.disconnect()

    execute_command = _raise_on_closed(redis.Redis.execute_command)
    transaction = _raise_on_closed(redis.Redis.transaction)
    pubsub = _raise_on_closed(redis.Redis.pubsub)


class UnknownExpire(enum.IntEnum):
    """Non-expiry (not ttls) results return from :func:`.get_expiry`.

    See: http://redis.io/commands/ttl or http://redis.io/commands/pttl
    """

    DOES_NOT_EXPIRE = -1
    """
    The command returns ``-1`` if the key exists but has no associated expire.
    """

    #: The command returns ``-2`` if the key does not exist.
    KEY_NOT_FOUND = -2


DOES_NOT_EXPIRE = UnknownExpire.DOES_NOT_EXPIRE
KEY_NOT_FOUND = UnknownExpire.KEY_NOT_FOUND

_UNKNOWN_EXPIRE_MAPPING = {e.value: e for e in list(UnknownExpire)}


def get_expiry(client, key, prior_version=None):
    """Gets an expiry for a key (using **best** determined ttl method)."""
    is_new_enough, _prior_version = is_server_new_enough(
        client, (2, 6), prior_version=prior_version)
    if is_new_enough:
        result = client.pttl(key)
        try:
            return _UNKNOWN_EXPIRE_MAPPING[result]
        except KeyError:
            return result / 1000.0
    else:
        result = client.ttl(key)
        try:
            return _UNKNOWN_EXPIRE_MAPPING[result]
        except KeyError:
            return float(result)


def apply_expiry(client, key, expiry, prior_version=None):
    """Applies an expiry to a key (using **best** determined expiry method)."""
    is_new_enough, _prior_version = is_server_new_enough(
        client, (2, 6), prior_version=prior_version)
    if is_new_enough:
        # Use milliseconds (as that is what pexpire uses/expects...)
        ms_expiry = expiry * 1000.0
        ms_expiry = max(0, int(ms_expiry))
        result = client.pexpire(key, ms_expiry)
    else:
        # Only supports seconds (not subseconds...)
        sec_expiry = int(expiry)
        sec_expiry = max(0, sec_expiry)
        result = client.expire(key, sec_expiry)
    return bool(result)


def is_server_new_enough(client, min_version,
                         default=False, prior_version=None):
    """Checks if a client is attached to a new enough redis server."""
    if not prior_version:
        try:
            server_info = client.info()
        except redis_exceptions.ResponseError:
            server_info = {}
        version_text = server_info.get('redis_version', '')
    else:
        version_text = prior_version
    version_pieces = []
    for p in version_text.split("."):
        try:
            version_pieces.append(int(p))
        except ValueError:
            break
    if not version_pieces:
        return (default, version_text)
    else:
        version_pieces = tuple(version_pieces)
        return (version_pieces >= min_version, version_text)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import reflection

from taskflow.engines.worker_based import endpoint as ep
from taskflow import task
from taskflow import test
from taskflow.tests import utils


class Task(task.Task):

    def __init__(self, a, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def execute(self, *args, **kwargs):
        pass


class TestEndpoint(test.TestCase):

    def setUp(self):
        super().setUp()
        self.task_cls = utils.TaskOneReturn
        self.task_uuid = 'task-uuid'
        self.task_args = {'context': 'context'}
        self.task_cls_name = reflection.get_class_name(self.task_cls)
        self.task_ep = ep.Endpoint(self.task_cls)
        self.task_result = 1

    def test_creation(self):
        task = self.task_ep.generate()
        self.assertEqual(self.task_cls_name, self.task_ep.name)
        self.assertIsInstance(task, self.task_cls)
        self.assertEqual(self.task_cls_name, task.name)

    def test_creation_with_task_name(self):
        task_name = 'test'
        task = self.task_ep.generate(name=task_name)
        self.assertEqual(self.task_cls_name, self.task_ep.name)
        self.assertIsInstance(task, self.task_cls)
        self.assertEqual(task_name, task.name)

    def test_creation_task_with_constructor_args(self):
        # NOTE(skudriashev): Exception is expected here since task
        # is created without any arguments passing to its constructor.
        endpoint = ep.Endpoint(Task)
        self.assertRaises(TypeError, endpoint.generate)

    def test_to_str(self):
        self.assertEqual(self.task_cls_name, str(self.task_ep))

    def test_execute(self):
        task = self.task_ep.generate(self.task_cls_name)
        result = self.task_ep.execute(task,
                                      task_uuid=self.task_uuid,
                                      arguments=self.task_args,
                                      progress_callback=None)
        self.assertEqual(self.task_result, result)

    def test_revert(self):
        task = self.task_ep.generate(self.task_cls_name)
        result = self.task_ep.revert(task,
                                     task_uuid=self.task_uuid,
                                     arguments=self.task_args,
                                     progress_callback=None,
                                     result=self.task_result,
                                     failures={})
        self.assertIsNone(result)



#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import futurist
import testtools

import taskflow.engines
from taskflow import exceptions as exc
from taskflow import test
from taskflow.tests import utils
from taskflow.utils import eventlet_utils as eu


class ArgumentsPassingTest(utils.EngineTestBase):

    def test_save_as(self):
        flow = utils.TaskOneReturn(name='task1', provides='first_data')
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({'first_data': 1}, engine.storage.fetch_all())

    def test_save_all_in_one(self):
        flow = utils.TaskMultiReturn(provides='all_data')
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({'all_data': (1, 3, 5)},
                         engine.storage.fetch_all())

    def test_save_several_values(self):
        flow = utils.TaskMultiReturn(provides=('badger', 'mushroom', 'snake'))
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({
            'badger': 1,
            'mushroom': 3,
            'snake': 5
        }, engine.storage.fetch_all())

    def test_save_dict(self):
        flow = utils.TaskMultiDict(provides={'badger',
                                             'mushroom',
                                             'snake'})
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({
            'badger': 0,
            'mushroom': 1,
            'snake': 2,
        }, engine.storage.fetch_all())

    def test_bad_save_as_value(self):
        self.assertRaises(TypeError,
                          utils.TaskOneReturn,
                          name='task1', provides=object())

    def test_arguments_passing(self):
        flow = utils.TaskMultiArgOneReturn(provides='result')
        engine = self._make_engine(flow)
        engine.storage.inject({'x': 1, 'y': 4, 'z': 9, 'a': 17})
        engine.run()
        self.assertEqual({
            'x': 1, 'y': 4, 'z': 9, 'a': 17,
            'result': 14,
        }, engine.storage.fetch_all())

    def test_arguments_missing(self):
        flow = utils.TaskMultiArg()
        engine = self._make_engine(flow)
        engine.storage.inject({'a': 1, 'b': 4, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)

    def test_partial_arguments_mapping(self):
        flow = utils.TaskMultiArgOneReturn(provides='result',
                                           rebind={'x': 'a'})
        engine = self._make_engine(flow)
        engine.storage.inject({'x': 1, 'y': 4, 'z': 9, 'a': 17})
        engine.run()
        self.assertEqual({
            'x': 1, 'y': 4, 'z': 9, 'a': 17,
            'result': 30,
        }, engine.storage.fetch_all())

    def test_argument_injection(self):
        flow = utils.TaskMultiArgOneReturn(provides='result',
                                           inject={'x': 1, 'y': 4, 'z': 9})
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({
            'result': 14,
        }, engine.storage.fetch_all())

    def test_argument_injection_rebind(self):
        flow = utils.TaskMultiArgOneReturn(provides='result',
                                           rebind=['a', 'b', 'c'],
                                           inject={'a': 1, 'b': 4, 'c': 9})
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({
            'result': 14,
        }, engine.storage.fetch_all())

    def test_argument_injection_required(self):
        flow = utils.TaskMultiArgOneReturn(provides='result',
                                           requires=['a', 'b', 'c'],
                                           inject={'x': 1, 'y': 4, 'z': 9,
                                                   'a': 0, 'b': 0, 'c': 0})
        engine = self._make_engine(flow)
        engine.run()
        self.assertEqual({
            'result': 14,
        }, engine.storage.fetch_all())

    def test_all_arguments_mapping(self):
        flow = utils.TaskMultiArgOneReturn(provides='result',
                                           rebind=['a', 'b', 'c'])
        engine = self._make_engine(flow)
        engine.storage.inject({
            'a': 1, 'b': 2, 'c': 3, 'x': 4, 'y': 5, 'z': 6
        })
        engine.run()
        self.assertEqual({
            'a': 1, 'b': 2, 'c': 3, 'x': 4, 'y': 5, 'z': 6,
            'result': 6,
        }, engine.storage.fetch_all())

    def test_invalid_argument_name_map(self):
        flow = utils.TaskMultiArg(rebind={'z': 'b'})
        engine = self._make_engine(flow)
        engine.storage.inject({'a': 1, 'y': 4, 'c': 9, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)

    def test_invalid_argument_name_list(self):
        flow = utils.TaskMultiArg(rebind=['a', 'z', 'b'])
        engine = self._make_engine(flow)
        engine.storage.inject({'a': 1, 'b': 4, 'c': 9, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)

    def test_bad_rebind_args_value(self):
        self.assertRaises(TypeError,
                          utils.TaskOneArg,
                          rebind=object())

    def test_long_arg_name(self):
        flow = utils.LongArgNameTask(requires='long_arg_name',
                                     provides='result')
        engine = self._make_engine(flow)
        engine.storage.inject({'long_arg_name': 1})
        engine.run()
        self.assertEqual({
            'long_arg_name': 1, 'result': 1
        }, engine.storage.fetch_all())

    def test_revert_rebound_args_required(self):
        flow = utils.TaskMultiArg(revert_rebind={'z': 'b'})
        engine = self._make_engine(flow)
        engine.storage.inject({'a': 1, 'y': 4, 'c': 9, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)

    def test_revert_required_args_required(self):
        flow = utils.TaskMultiArg(revert_requires=['a'])
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 4, 'z': 9, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)

    def test_derived_revert_args_required(self):
        flow = utils.TaskRevertExtraArgs()
        engine = self._make_engine(flow)
        engine.storage.inject({'y': 4, 'z': 9, 'x': 17})
        self.assertRaises(exc.MissingDependencies, engine.run)
        engine.storage.inject({'revert_arg': None})
        self.assertRaises(exc.ExecutionFailure, engine.run)


class SerialEngineTest(ArgumentsPassingTest, test.TestCase):

    def _make_engine(self, flow, flow_detail=None):
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     engine='serial',
                                     backend=self.backend)


class ParallelEngineWithThreadsTest(ArgumentsPassingTest, test.TestCase):
    _EXECUTOR_WORKERS = 2

    def _make_engine(self, flow, flow_detail=None, executor=None):
        if executor is None:
            executor = 'threads'
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     engine='parallel',
                                     backend=self.backend,
                                     executor=executor,
                                     max_workers=self._EXECUTOR_WORKERS)


@testtools.skipIf(not eu.EVENTLET_AVAILABLE, 'eventlet is not available')
class ParallelEngineWithEventletTest(ArgumentsPassingTest, test.TestCase):

    def _make_engine(self, flow, flow_detail=None, executor=None):
        if executor is None:
            executor = futurist.GreenThreadPoolExecutor()
            self.addCleanup(executor.shutdown)
        return taskflow.engines.load(flow,
                                     flow_detail=flow_detail,
                                     backend=self.backend,
                                     engine='parallel',
                                     executor=executor)



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Switch postgres to json native type.

Revision ID: 2ad4984f2864
Revises: 3162c0f3f8e4
Create Date: 2015-06-04 13:08:36.667948

"""

# revision identifiers, used by Alembic.
revision = '2ad4984f2864'
down_revision = '3162c0f3f8e4'

from alembic import op


_ALTER_TO_JSON_TPL = 'ALTER TABLE %s ALTER COLUMN %s TYPE JSON USING %s::JSON'
_TABLES_COLS = tuple([
    ('logbooks', 'meta'),
    ('flowdetails', 'meta'),
    ('atomdetails', 'meta'),
    ('atomdetails', 'failure'),
    ('atomdetails', 'revert_failure'),
    ('atomdetails', 'results'),
    ('atomdetails', 'revert_results'),
])
_ALTER_TO_TEXT_TPL = 'ALTER TABLE %s ALTER COLUMN %s TYPE TEXT'


def upgrade():
    b = op.get_bind()
    if b.dialect.name.startswith('postgresql'):
        for (table_name, col_name) in _TABLES_COLS:
            q = _ALTER_TO_JSON_TPL % (table_name, col_name, col_name)
            op.execute(q)


def downgrade():
    b = op.get_bind()
    if b.dialect.name.startswith('postgresql'):
        for (table_name, col_name) in _TABLES_COLS:
            q = _ALTER_TO_TEXT_TPL % (table_name, col_name)
            op.execute(q)



#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow import flow
from taskflow.types import graph as gr


class Flow(flow.Flow):
    """Unordered flow pattern.

    A unordered (potentially nested) flow of *tasks/flows* that can be
    executed in any order as one unit and rolled back as one unit.
    """

    def __init__(self, name, retry=None):
        super().__init__(name, retry)
        self._graph = gr.Graph(name=name)

    def add(self, *items):
        """Adds a given task/tasks/flow/flows to this flow."""
        for item in items:
            if not self._graph.has_node(item):
                self._graph.add_node(item)
        return self

    def __len__(self):
        return len(self._graph)

    def __iter__(self):
        yield from self._graph

    def iter_links(self):
        yield from self._graph.edges(data=True)

    def iter_nodes(self):
        yield from self._graph.nodes(data=True)

    @property
    def requires(self):
        requires = set()
        retry_provides = set()
        if self._retry is not None:
            requires.update(self._retry.requires)
            retry_provides.update(self._retry.provides)
        for item in self:
            item_requires = item.requires - retry_provides
            requires.update(item_requires)
        return frozenset(requires)



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from taskflow import exceptions as exc

# Job states.
CLAIMED = 'CLAIMED'
COMPLETE = 'COMPLETE'
UNCLAIMED = 'UNCLAIMED'

# Flow states.
FAILURE = 'FAILURE'
PENDING = 'PENDING'
REVERTING = 'REVERTING'
REVERTED = 'REVERTED'
RUNNING = 'RUNNING'
SUCCESS = 'SUCCESS'
SUSPENDING = 'SUSPENDING'
SUSPENDED = 'SUSPENDED'
RESUMING = 'RESUMING'

# Task states (mainly a subset of the flow states).
FAILURE = FAILURE
PENDING = PENDING
REVERTED = REVERTED
REVERTING = REVERTING
SUCCESS = SUCCESS
RUNNING = RUNNING
RETRYING = 'RETRYING'
IGNORE = 'IGNORE'
REVERT_FAILURE = 'REVERT_FAILURE'

# Atom intentions.
EXECUTE = 'EXECUTE'
IGNORE = IGNORE
REVERT = 'REVERT'
RETRY = 'RETRY'
INTENTIONS = (EXECUTE, IGNORE, REVERT, RETRY)

# Additional engine states
SCHEDULING = 'SCHEDULING'
WAITING = 'WAITING'
ANALYZING = 'ANALYZING'

# Job state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html

_ALLOWED_JOB_TRANSITIONS = frozenset((
    # Job is being claimed.
    (UNCLAIMED, CLAIMED),

    # Job has been lost (or manually unclaimed/abandoned).
    (CLAIMED, UNCLAIMED),

    # Job has been finished.
    (CLAIMED, COMPLETE),
))


def check_job_transition(old_state, new_state):
    """Check that job can transition from from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true. If transition
    should be ignored, it returns false. If transition is not
    valid, it raises an InvalidState exception.
    """
    if old_state == new_state:
        return False
    pair = (old_state, new_state)
    if pair in _ALLOWED_JOB_TRANSITIONS:
        return True
    raise exc.InvalidState("Job transition from '%s' to '%s' is not allowed"
                           % pair)


# Flow state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#flow

_ALLOWED_FLOW_TRANSITIONS = frozenset((
    (PENDING, RUNNING),       # run it!

    (RUNNING, SUCCESS),       # all tasks finished successfully
    (RUNNING, FAILURE),       # some of task failed
    (RUNNING, REVERTED),      # some of task failed and flow has been reverted
    (RUNNING, SUSPENDING),    # engine.suspend was called
    (RUNNING, RESUMING),      # resuming from a previous running

    (SUCCESS, RUNNING),       # see note below

    (FAILURE, RUNNING),       # see note below

    (REVERTED, PENDING),      # try again
    (SUCCESS, PENDING),       # run it again

    (SUSPENDING, SUSPENDED),  # suspend finished
    (SUSPENDING, SUCCESS),    # all tasks finished while we were waiting
    (SUSPENDING, FAILURE),    # some tasks failed while we were waiting
    (SUSPENDING, REVERTED),   # all tasks were reverted while we were waiting
    (SUSPENDING, RESUMING),   # resuming from a previous suspending

    (SUSPENDED, RUNNING),     # restart from suspended

    (RESUMING, SUSPENDED),    # after flow resumed, it is suspended
))


# NOTE(imelnikov) SUCCESS->RUNNING and FAILURE->RUNNING transitions are
# useful when flow or flowdetails backing it were altered after the flow
# was finished; then, client code may want to run through flow again
# to ensure all tasks from updated flow had a chance to run.


# NOTE(imelnikov): Engine cannot transition flow from SUSPENDING to
# SUSPENDED while some tasks from the flow are running and some results
# from them are not retrieved and saved properly, so while flow is
# in SUSPENDING state it may wait for some of the tasks to stop. Then,
# flow can go to SUSPENDED, SUCCESS, FAILURE or REVERTED state depending
# of actual state of the tasks -- e.g. if all tasks were finished
# successfully while we were waiting, flow can be transitioned from
# SUSPENDING to SUCCESS state.

_IGNORED_FLOW_TRANSITIONS = frozenset(
    (a, b)
    for a in (PENDING, FAILURE, SUCCESS, SUSPENDED, REVERTED)
    for b in (SUSPENDING, SUSPENDED, RESUMING)
    if a != b
)


def check_flow_transition(old_state, new_state):
    """Check that flow can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true. If transition
    should be ignored, it returns false. If transition is not
    valid, it raises an InvalidState exception.
    """
    if old_state == new_state:
        return False
    pair = (old_state, new_state)
    if pair in _ALLOWED_FLOW_TRANSITIONS:
        return True
    if pair in _IGNORED_FLOW_TRANSITIONS:
        return False
    raise exc.InvalidState("Flow transition from '%s' to '%s' is not allowed"
                           % pair)


# Task state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#task

_ALLOWED_TASK_TRANSITIONS = frozenset((
    (PENDING, RUNNING),       # run it!
    (PENDING, IGNORE),        # skip it!

    (RUNNING, SUCCESS),       # the task executed successfully
    (RUNNING, FAILURE),       # the task execution failed

    (FAILURE, REVERTING),     # task execution failed, try reverting...
    (SUCCESS, REVERTING),     # some other task failed, try reverting...

    (REVERTING, REVERTED),           # the task reverted successfully
    (REVERTING, REVERT_FAILURE),     # the task failed reverting (terminal!)

    (REVERTED, PENDING),      # try again
    (IGNORE, PENDING),        # try again
))


def check_task_transition(old_state, new_state):
    """Check that task can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true, false otherwise.
    """
    pair = (old_state, new_state)
    if pair in _ALLOWED_TASK_TRANSITIONS:
        return True
    return False


# Retry state transitions
# See: https://docs.openstack.org/taskflow/latest/user/states.html#retry

_ALLOWED_RETRY_TRANSITIONS = list(_ALLOWED_TASK_TRANSITIONS)
_ALLOWED_RETRY_TRANSITIONS.extend([
    (SUCCESS, RETRYING),      # retrying retry controller
    (RETRYING, RUNNING),      # run retry controller that has been retrying
])
_ALLOWED_RETRY_TRANSITIONS = frozenset(_ALLOWED_RETRY_TRANSITIONS)


def check_retry_transition(old_state, new_state):
    """Check that retry can transition from ``old_state`` to ``new_state``.

    If transition can be performed, it returns true, false otherwise.
    """
    pair = (old_state, new_state)
    if pair in _ALLOWED_RETRY_TRANSITIONS:
        return True
    return False



#    Copyright (C) Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import testtools

from taskflow import exceptions as exc
from taskflow.jobs.backends import impl_etcd
from taskflow.jobs import base as jobs_base
from taskflow import test
from taskflow.tests.unit.jobs import base
from taskflow.tests import utils as test_utils

ETCD_AVAILABLE = test_utils.etcd_available()
ETCD_PORT = test_utils.ETCD_PORT


class EtcdJobBoardMixin:
    def create_board(self, conf=None, persistence=None):
        self.path = f"test-{uuidutils.generate_uuid()}"
        board_conf = {
            "port": ETCD_PORT,
            "path": self.path,
        }
        if conf:
            board_conf.update(conf)
        board = impl_etcd.EtcdJobBoard("etcd", board_conf,
                                       persistence=persistence)
        return board._client, board


class MockedEtcdJobBoard(test.TestCase, EtcdJobBoardMixin):

    def test_create_board(self):
        _, jobboard = self.create_board()
        self.assertEqual(f"/taskflow/jobs/{self.path}", jobboard._root_path)

        _, jobboard = self.create_board({"path": "/testpath"})
        self.assertEqual("/taskflow/jobs/testpath", jobboard._root_path)

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard.incr")
    @mock.patch("threading.Condition")
    @mock.patch("oslo_utils.uuidutils.generate_uuid")
    @mock.patch("oslo_utils.timeutils.utcnow")
    def test_post(self,
                  mock_utcnow: mock.Mock,
                  mock_generated_uuid: mock.Mock,
                  mock_cond: mock.Mock,
                  mock_incr: mock.Mock):
        mock_incr.return_value = 12
        mock_generated_uuid.return_value = "uuid1"
        mock_utcnow.return_value = "utcnow1"

        mock_book = mock.Mock()
        mock_book.name = "book1_name"
        mock_book.uuid = "book1_uuid"
        mock_details = mock.Mock()

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()
        job = jobboard.post("post1", book=mock_book,
                            details=mock_details,
                            priority=jobs_base.JobPriority.NORMAL)

        expected_key = (
            f"/taskflow/jobs/{self.path}/job12")
        expected_data_key = expected_key + jobboard.DATA_POSTFIX
        expected_book_data = {
            "name": "book1_name",
            "uuid": "book1_uuid"
        }
        expected_job_posting = {
            "uuid": "uuid1",
            "name": "post1",
            "priority": "NORMAL",
            "created_on": "utcnow1",
            "details": mock_details,
            "book": expected_book_data,
            "sequence": 12,
        }

        mock_incr.assert_called_with(f"/taskflow/jobs/{self.path}/sequence")

        jobboard._client.create.assert_called_with(
            expected_data_key, jsonutils.dumps(expected_job_posting))

        self.assertEqual("post1", job.name)
        self.assertEqual(expected_key, job.key)
        self.assertEqual(mock_details, job.details)
        self.assertEqual(mock_book, job.book)
        self.assertEqual(expected_book_data, job._book_data)
        self.assertEqual(jobs_base.JobPriority.NORMAL, job.priority)
        self.assertEqual(12, job.sequence)

        self.assertEqual(1, len(jobboard._job_cache))
        self.assertEqual(job, jobboard._job_cache[expected_key])

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "set_last_modified")
    def test_claim(self, mock_set_last_modified):
        who = "owner1"
        lease_id = uuidutils.generate_uuid()

        _, jobboard = self.create_board(conf={"ttl": 37})
        jobboard._client = mock.Mock()

        mock_lease = mock.Mock(id=lease_id)
        jobboard._client.lease.return_value = mock_lease
        jobboard._client.create.return_value = True
        jobboard._client.get.return_value = [mock.Mock()]

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7",
                                uuid=uuidutils.generate_uuid(),
                                details=mock.Mock(),
                                backend="etcd",
                                book=mock.Mock(),
                                book_data=mock.Mock(),
                                priority=jobs_base.JobPriority.NORMAL,
                                sequence=7,
                                created_on="date")

        jobboard.claim(job, who)

        jobboard._client.lease.assert_called_once_with(ttl=37)

        jobboard._client.create.assert_called_once_with(
            f"{job.key}{jobboard.LOCK_POSTFIX}",
            jsonutils.dumps({"owner": who,
                             "lease_id": lease_id}),
            lease=mock_lease)

        jobboard._client.get.assert_called_once_with(
            job.key + jobboard.DATA_POSTFIX)
        mock_lease.revoke.assert_not_called()

        mock_set_last_modified.assert_called_once_with(job)

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "set_last_modified")
    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "find_owner")
    def test_claim_already_claimed(self, mock_find_owner,
                                   mock_set_last_modified):
        who = "owner1"
        lease_id = uuidutils.generate_uuid()

        mock_find_owner.return_value = who

        _, jobboard = self.create_board({"ttl": 37})
        jobboard._client = mock.Mock()

        mock_lease = mock.Mock(id=lease_id)
        jobboard._client.lease.return_value = mock_lease
        jobboard._client.create.return_value = False
        jobboard._client.get.return_value = []

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7",
                                uuid=uuidutils.generate_uuid(),
                                details=mock.Mock(),
                                backend="etcd",
                                book=mock.Mock(),
                                book_data=mock.Mock(),
                                priority=jobs_base.JobPriority.NORMAL,
                                sequence=7,
                                created_on="date")

        self.assertRaisesRegex(exc.UnclaimableJob, "already claimed by",
                               jobboard.claim, job, who)

        jobboard._client.lease.assert_called_once_with(ttl=37)

        jobboard._client.create.assert_called_once_with(
            f"{job.key}{jobboard.LOCK_POSTFIX}",
            jsonutils.dumps({"owner": who,
                             "lease_id": lease_id}),
            lease=mock_lease)

        mock_lease.revoke.assert_called_once()

        mock_set_last_modified.assert_not_called()

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "set_last_modified")
    def test_claim_deleted(self, mock_set_last_modified):
        who = "owner1"
        lease_id = uuidutils.generate_uuid()

        _, jobboard = self.create_board({"ttl": 37})
        jobboard._client = mock.Mock()

        mock_lease = mock.Mock(id=lease_id)
        jobboard._client.lease.return_value = mock_lease
        jobboard._client.create.return_value = True
        jobboard._client.get.return_value = []

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7",
                                uuid=uuidutils.generate_uuid(),
                                details=mock.Mock(),
                                backend="etcd",
                                book=mock.Mock(),
                                book_data=mock.Mock(),
                                priority=jobs_base.JobPriority.NORMAL,
                                sequence=7,
                                created_on="date")

        self.assertRaisesRegex(exc.UnclaimableJob, "already deleted",
                               jobboard.claim, job, who)

        jobboard._client.lease.assert_called_once_with(ttl=37)

        jobboard._client.create.assert_called_once_with(
            f"{job.key}{jobboard.LOCK_POSTFIX}",
            jsonutils.dumps({"owner": who,
                             "lease_id": lease_id}),
            lease=mock_lease)

        jobboard._client.get.assert_called_once_with(
            job.key + jobboard.DATA_POSTFIX)
        mock_lease.revoke.assert_called_once()

        mock_set_last_modified.assert_not_called()

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "_remove_job_from_cache")
    def test_consume(self, mock__remove_job_from_cache,
                     mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner1", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        jobboard.consume(job, "owner1")

        jobboard._client.delete_prefix.assert_called_once_with(job.key + ".")
        mock__remove_job_from_cache.assert_called_once_with(job.key)

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    def test_consume_bad_owner(self, mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner2", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        self.assertRaisesRegex(exc.JobFailure, "which is not owned",
                               jobboard.consume, job, "owner1")

        jobboard._client.delete_prefix.assert_not_called()

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    def test_abandon(self, mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner1", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        jobboard.abandon(job, "owner1")

        jobboard._client.delete.assert_called_once_with(
            f"{job.key}{jobboard.LOCK_POSTFIX}")

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    def test_abandon_bad_owner(self, mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner2", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        self.assertRaisesRegex(exc.JobFailure, "which is not owned",
                               jobboard.abandon, job, "owner1")

        jobboard._client.delete.assert_not_called()

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "_remove_job_from_cache")
    def test_trash(self, mock__remove_job_from_cache,
                   mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner1", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        jobboard.trash(job, "owner1")

        jobboard._client.create.assert_called_once_with(
            f"/taskflow/.trash/{self.path}/job7", mock.ANY)
        jobboard._client.delete_prefix.assert_called_once_with(job.key + ".")
        mock__remove_job_from_cache.assert_called_once_with(job.key)

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "_remove_job_from_cache")
    def test_trash_bad_owner(self, mock__remove_job_from_cache,
                             mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner2", mock.Mock()]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        self.assertRaisesRegex(exc.JobFailure, "which is not owned",
                               jobboard.trash, job, "owner1")

        jobboard._client.create.assert_not_called()
        jobboard._client.delete_prefix.assert_not_called()
        mock__remove_job_from_cache.assert_not_called()

    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "get_owner_and_data")
    @mock.patch("taskflow.jobs.backends.impl_etcd.EtcdJobBoard."
                "_remove_job_from_cache")
    def test_trash_deleted_job(self, mock__remove_job_from_cache,
                               mock_get_owner_and_data):
        mock_get_owner_and_data.return_value = ["owner1", None]

        _, jobboard = self.create_board()
        jobboard._client = mock.Mock()

        job = impl_etcd.EtcdJob(jobboard,
                                "job7",
                                jobboard._client,
                                f"/taskflow/jobs/{self.path}/job7")
        self.assertRaisesRegex(exc.NotFound, "Cannot find job",
                               jobboard.trash, job, "owner1")

        jobboard._client.create.assert_not_called()
        jobboard._client.delete_prefix.assert_not_called()
        mock__remove_job_from_cache.assert_not_called()


@testtools.skipIf(not ETCD_AVAILABLE, 'Etcd is not available')
class EtcdJobBoardTest(test.TestCase, base.BoardTestMixin, EtcdJobBoardMixin):
    def setUp(self):
        super().setUp()
        self.client, self.board = self.create_board()

    def test__incr(self):
        key = uuidutils.generate_uuid()

        self.board.connect()
        self.addCleanup(self.board.close)
        self.addCleanup(self.board._client.delete, key)

        self.assertEqual(1, self.board.incr(key))
        self.assertEqual(2, self.board.incr(key))
        self.assertEqual(3, self.board.incr(key))

        self.assertEqual(b'3', self.board.get_one(key))
        self.board.close()

    def test_get_one(self):
        key1 = uuidutils.generate_uuid()

        self.board.connect()
        self.addCleanup(self.board._client.delete, key1)

        # put data and get it
        self.board._client.put(key1, "testset1")
        self.assertEqual(b"testset1", self.board.get_one(key1))

        # delete data and check that it's not found
        self.board._client.delete(key1)
        self.assertIsNone(self.board.get_one(key1))

        # get a non-existant data
        key2 = uuidutils.generate_uuid()
        # (ensure it doesn't exist)
        self.board._client.delete(key2)
        self.assertIsNone(self.board.get_one(key2))
        self.board.close()



#    Copyright (C) Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
import typing

import etcd3gw
import fasteners
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

from taskflow import exceptions as exc
from taskflow.jobs import base
from taskflow import logging
from taskflow import states
from taskflow.utils import misc
if typing.TYPE_CHECKING:
    from taskflow.types import entity

LOG = logging.getLogger(__name__)


class EtcdJob(base.Job):
    """An Etcd job."""

    board: 'EtcdJobBoard'

    def __init__(self, board: 'EtcdJobBoard', name, client, key,
                 uuid=None, details=None, backend=None,
                 book=None, book_data=None,
                 priority=base.JobPriority.NORMAL,
                 sequence=None, created_on=None):
        super().__init__(board, name, uuid=uuid, details=details,
                         backend=backend, book=book, book_data=book_data)

        self._client = client
        self._key = key
        self._priority = priority
        self._sequence = sequence
        self._created_on = created_on
        self._root = board._root_path

        self._lease = None

    @property
    def key(self):
        return self._key

    @property
    def last_modified(self):
        try:
            raw_data = self.board.get_last_modified(self)
            data = jsonutils.loads(raw_data)
            ret = timeutils.parse_strtime(data["last_modified"])
            return ret
        except Exception:
            LOG.exception("Cannot read load_modified key.")
        return 0

    @property
    def created_on(self):
        return self._created_on

    @property
    def state(self):
        """Access the current state of this job."""
        owner, data = self.board.get_owner_and_data(self)
        if not data:
            if owner is not None:
                LOG.info(f"Owner key was found for job {self.uuid}, "
                         f"but the key {self.key} is missing")
            return states.COMPLETE
        if not owner:
            return states.UNCLAIMED
        return states.CLAIMED

    @property
    def sequence(self):
        return self._sequence

    @property
    def priority(self):
        return self._priority

    @property
    def lease(self):
        if not self._lease:
            owner_data = self.board.get_owner_data(self)
            if 'lease_id' not in owner_data:
                return None
            lease_id = owner_data['lease_id']
            self._lease = etcd3gw.Lease(id=lease_id,
                                        client=self._client)
        return self._lease

    def expires_in(self):
        """How many seconds until the claim expires."""
        if self.lease is None:
            return -1
        return self.lease.ttl()

    def extend_expiry(self, expiry):
        """Extends the owner key (aka the claim) expiry for this job.

        Returns ``True`` if the expiry request was performed
        otherwise ``False``.
        """
        if self.lease is None:
            return False
        ret = self.lease.refresh()
        return (ret > 0)

    @property
    def root(self):
        return self._root

    def __lt__(self, other):
        if not isinstance(other, EtcdJob):
            return NotImplemented
        if self.root == other.root:
            if self.priority == other.priority:
                return self.sequence < other.sequence
            else:
                ordered = base.JobPriority.reorder(
                    (self.priority, self), (other.priority, other))
                if ordered[0] is self:
                    return False
                return True
        else:
            # Different jobboards with different roots...
            return self.root < other.root

    def __eq__(self, other):
        if not isinstance(other, EtcdJob):
            return NotImplemented
        return ((self.root, self.sequence, self.priority) ==
                (other.root, other.sequence, other.priority))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.key)


class EtcdJobBoard(base.JobBoard):
    """A jobboard backed by `etcd`_.

    This jobboard creates sequenced key/value pairs in etcd. Each key
    represents a job and its associated value contains the parameter of the
    job encoded in
    json.
    The users of the jobboard can iterate over the available job and decide if
    they want to attempt to claim one job by calling the :meth:`.claim` method.
    Claiming a job consists in atomically create a key based on the key of job
    and the ".lock" postfix. If the atomic creation of the key is successful
    the job belongs to the user. Any attempt to lock an already locked job
    will fail.
    When a job is complete, the user consumes the job by calling the
    :meth:`.consume` method, it deletes the job and the lock from etcd.
    Alternatively, a user can trash (:meth:`.trash`) or abandon
    (:meth:`.abandon`) if they want to delete the job or leave it for another
    user.
    Etcd doesn't provide a method for unlocking the jobs when a consumer dies.
    The Etcd jobboard provides timed expirations, based on a global ``ttl``
    configuration setting or the ``expiry`` parameter of the :meth:`.claim`
    method. When this time-to-live/expiry is reached, the job is automatically
    unlocked and another consumer can claim it. If it is expected that a task
    of a job takes more time than the defined time-to-live, the
    consumer can refresh the timer by calling the :meth:`EtcdJob.extend_expiry`
    function.

    .. _etcd: https://etcd.io/
    """
    ROOT_PATH = "/taskflow/jobs"

    TRASH_PATH = "/taskflow/.trash"

    DEFAULT_PATH = "jobboard"

    JOB_PREFIX = "job"

    SEQUENCE_KEY = "sequence"

    DATA_POSTFIX = ".data"

    LOCK_POSTFIX = ".lock"

    LAST_MODIFIED_POSTFIX = ".last_modified"

    ETCD_CONFIG_OPTIONS = (
        ("host", str),
        ("port", int),
        ("protocol", str),
        ("ca_cert", str),
        ("cert_key", str),
        ("cert_cert", str),
        ("timeout", float),
        ("api_path", str),
    )

    INIT_STATE = 'init'
    CONNECTED_STATE = 'connected'
    FETCH_STATE = 'fetched'

    _client: etcd3gw.Etcd3Client

    def __init__(self, name, conf, client=None, persistence=None):
        super().__init__(name, conf)

        self._client = client
        self._persistence = persistence
        self._state = self.INIT_STATE

        path_elems = [self.ROOT_PATH,
                      self._conf.get("path", self.DEFAULT_PATH)]
        self._root_path = self._create_path(*path_elems)

        self._job_cache = {}
        self._job_cond = threading.Condition()

        self._open_close_lock = threading.RLock()

        self._watcher_thd = None
        self._thread_cancel = None
        self._watcher = None
        self._watcher_cancel = None

    def _create_path(self, root, *args):
        return "/".join([root] + [a.strip("/") for a in args])

    def incr(self, key):
        """Atomically increment an integer, create it if it doesn't exist"""
        while True:
            value = self._client.get(key)
            if not value:
                res = self._client.create(key, 1)
                if res:
                    return 1
                # Another thread has just created the key after we failed to
                # read it, retry to get the new current value
                continue

            value = int(value[0])
            next_value = value + 1

            res = self._client.replace(key, value, next_value)
            if res:
                return next_value

    def get_one(self, key):
        if self._client is None:
            raise exc.JobFailure(f"Cannot read key {key}, client is closed")
        value = self._client.get(key)
        if not value:
            return None
        return value[0]

    def _fetch_jobs(self, only_unclaimed=False, ensure_fresh=False):
        # TODO(gthiemonge) only_unclaimed is ignored
        if ensure_fresh or self._state != self.FETCH_STATE:
            self._ensure_fresh()
        return sorted(self._job_cache.values())

    def _ensure_fresh(self):
        prefix = self._create_path(self._root_path, self.JOB_PREFIX)
        jobs = self._client.get_prefix(prefix)
        listed_jobs = {}
        for job in jobs:
            data, metadata = job
            key = misc.binary_decode(metadata['key'])
            if key.endswith(self.DATA_POSTFIX):
                key = key.rstrip(self.DATA_POSTFIX)
                listed_jobs[key] = data

        removed_jobs = []
        with self._job_cond:
            for key in self._job_cache.keys():
                if key not in listed_jobs:
                    removed_jobs.append(key)
        for key in removed_jobs:
            self._remove_job_from_cache(key)

        for key, data in listed_jobs.items():
            self._process_incoming_job(key, data)
        self._state = self.FETCH_STATE

    def _process_incoming_job(self, key, data):
        try:
            job_data = jsonutils.loads(data)
        except jsonutils.json.JSONDecodeError:
            msg = ("Incorrectly formatted job data found at "
                   f"key: {key}")
            LOG.warning(msg, exc_info=True)
            LOG.info("Deleting invalid job data at key: %s", key)
            self._client.delete(key)
            raise exc.JobFailure(msg)

        with self._job_cond:
            if key not in self._job_cache:
                job_priority = base.JobPriority.convert(job_data["priority"])
                new_job = EtcdJob(self,
                                  job_data["name"],
                                  self._client,
                                  key,
                                  uuid=job_data["uuid"],
                                  details=job_data.get("details", {}),
                                  backend=self._persistence,
                                  book_data=job_data.get("book"),
                                  priority=job_priority,
                                  sequence=job_data["sequence"])
                self._job_cache[key] = new_job
                self._job_cond.notify_all()

    def _remove_job_from_cache(self, key):
        """Remove job from cache."""
        with self._job_cond:
            if key in self._job_cache:
                self._job_cache.pop(key, None)

    def _board_removal_func(self, job):
        try:
            self._remove_job_from_cache(job.key)
            self._client.delete_prefix(job.key)
        except Exception:
            LOG.exception(f"Failed to delete prefix {job.key}")

    def iterjobs(self, only_unclaimed=False, ensure_fresh=False):
        """Returns an iterator of jobs that are currently on this board."""
        return base.JobBoardIterator(
            self, LOG, only_unclaimed=only_unclaimed,
            ensure_fresh=ensure_fresh,
            board_fetch_func=self._fetch_jobs,
            board_removal_func=self._board_removal_func)

    def wait(self, timeout=None):
        """Waits a given amount of time for **any** jobs to be posted."""
        # Wait until timeout expires (or forever) for jobs to appear.
        watch = timeutils.StopWatch(duration=timeout)
        watch.start()
        with self._job_cond:
            while True:
                if not self._job_cache:
                    if watch.expired():
                        raise exc.NotFound("Expired waiting for jobs to"
                                           " arrive; waited %s seconds"
                                           % watch.elapsed())
                    # This is done since the given timeout can not be provided
                    # to the condition variable, since we can not ensure that
                    # when we acquire the condition that there will actually
                    # be jobs (especially if we are spuriously awaken), so we
                    # must recalculate the amount of time we really have left.
                    self._job_cond.wait(watch.leftover(return_none=True))
                else:
                    curr_jobs = self._fetch_jobs()
                    fetch_func = lambda ensure_fresh: curr_jobs
                    removal_func = lambda a_job: self._remove_job_from_cache(
                        a_job.key)
                    return base.JobBoardIterator(
                        self, LOG, board_fetch_func=fetch_func,
                        board_removal_func=removal_func)

    @property
    def job_count(self):
        """Returns how many jobs are on this jobboard."""
        return len(self._job_cache)

    def get_owner_data(self, job: EtcdJob) -> dict | None:
        owner_key = job.key + self.LOCK_POSTFIX
        owner_data = self.get_one(owner_key)
        if not owner_data:
            return None
        return jsonutils.loads(owner_data)

    def find_owner(self, job: EtcdJob) -> dict | None:
        """Gets the owner of the job if one exists."""
        data = self.get_owner_data(job)
        if data:
            return data['owner']
        return None

    def get_data(self, job: EtcdJob) -> bytes:
        key = job.key + self.DATA_POSTFIX
        return self.get_one(key)

    def get_owner_and_data(self, job: EtcdJob) -> tuple[
            str | None, bytes | None]:
        if self._client is None:
            raise exc.JobFailure("Cannot retrieve information, "
                                 "not connected")

        job_data = None
        job_owner = None

        for data, metadata in self._client.get_prefix(job.key + "."):
            key = misc.binary_decode(metadata["key"])
            if key.endswith(self.DATA_POSTFIX):
                # bytes?
                job_data = data
            elif key.endswith(self.LOCK_POSTFIX):
                data = jsonutils.loads(data)
                job_owner = data["owner"]

        return job_owner, job_data

    def set_last_modified(self, job: EtcdJob):
        key = job.key + self.LAST_MODIFIED_POSTFIX

        now = timeutils.utcnow()
        self._client.put(key, jsonutils.dumps({"last_modified": now}))

    def get_last_modified(self, job: EtcdJob):
        key = job.key + self.LAST_MODIFIED_POSTFIX

        return self.get_one(key)

    def post(self, name, book=None, details=None,
             priority=base.JobPriority.NORMAL) -> EtcdJob:
        """Atomically creates and posts a job to the jobboard."""
        job_priority = base.JobPriority.convert(priority)
        job_uuid = uuidutils.generate_uuid()
        job_posting = base.format_posting(job_uuid, name,
                                          created_on=timeutils.utcnow(),
                                          book=book, details=details,
                                          priority=job_priority)
        seq = self.incr(self._create_path(self._root_path, self.SEQUENCE_KEY))
        key = self._create_path(self._root_path, f"{self.JOB_PREFIX}{seq}")

        job_posting["sequence"] = seq
        raw_job_posting = jsonutils.dumps(job_posting)

        data_key = key + self.DATA_POSTFIX

        self._client.create(data_key, raw_job_posting)
        job = EtcdJob(self, name, self._client, key,
                      uuid=job_uuid,
                      details=details,
                      backend=self._persistence,
                      book=book,
                      book_data=job_posting.get('book'),
                      priority=job_priority,
                      sequence=seq)
        with self._job_cond:
            self._job_cache[key] = job
            self._job_cond.notify_all()
        return job

    @base.check_who
    def claim(self, job, who, expiry=None):
        """Atomically attempts to claim the provided job."""
        owner_key = job.key + self.LOCK_POSTFIX

        ttl = expiry or self._conf.get('ttl', None)

        if ttl:
            lease = self._client.lease(ttl=ttl)
        else:
            lease = None

        owner_dict = {
            "owner": who,
        }
        if lease:
            owner_dict["lease_id"] = lease.id

        owner_value = jsonutils.dumps(owner_dict)

        # Create a lock for the job, if the lock already exists, the job
        # is owned by another worker
        created = self._client.create(owner_key, owner_value, lease=lease)
        if not created:
            # Creation is denied, revoke the lease, we cannot claim the job.
            if lease:
                lease.revoke()

            owner = self.find_owner(job)
            if owner:
                message = f"Job {job.uuid} already claimed by '{owner}'"
            else:
                message = f"Job {job.uuid} already claimed"
            raise exc.UnclaimableJob(message)

        # Ensure that the job still exists, it may have been claimed and
        # consumed by another thread before we enter this function
        if not self.get_data(job):
            # Revoke the lease
            if lease:
                lease.revoke()
            else:
                self._client.delete(owner_key)
            raise exc.UnclaimableJob(f"Job {job.uuid} already deleted.")

        self.set_last_modified(job)

    @base.check_who
    def consume(self, job, who):
        """Permanently (and atomically) removes a job from the jobboard."""
        owner, data = self.get_owner_and_data(job)
        if data is None or owner is None:
            raise exc.NotFound(f"Cannot find job {job.uuid}")
        if owner != who:
            raise exc.JobFailure(f"Cannot consume a job {job.uuid}"
                                 f" which is not owned by {who}")

        self._client.delete_prefix(job.key + ".")
        self._remove_job_from_cache(job.key)

    @base.check_who
    def abandon(self, job, who):
        """Atomically attempts to abandon the provided job."""
        owner, data = self.get_owner_and_data(job)
        if data is None or owner is None:
            raise exc.NotFound(f"Cannot find job {job.uuid}")
        if owner != who:
            raise exc.JobFailure(f"Cannot abandon a job {job.uuid}"
                                 f" which is not owned by {who}")

        owner_key = job.key + self.LOCK_POSTFIX
        self._client.delete(owner_key)

    @base.check_who
    def trash(self, job, who):
        """Trash the provided job."""
        owner, data = self.get_owner_and_data(job)
        if data is None or owner is None:
            raise exc.NotFound(f"Cannot find job {job.uuid}")
        if owner != who:
            raise exc.JobFailure(f"Cannot trash a job {job.uuid} "
                                 f"which is not owned by {who}")

        trash_key = job.key.replace(self.ROOT_PATH, self.TRASH_PATH)
        self._client.create(trash_key, data)
        self._client.delete_prefix(job.key + ".")
        self._remove_job_from_cache(job.key)

    def register_entity(self, entity: 'entity.Entity'):
        """Register an entity to the jobboard('s backend), e.g: a conductor"""
        # TODO(gthiemonge) Doesn't seem to be useful with Etcd

    @property
    def connected(self):
        """Returns if this jobboard is connected."""
        return self._client is not None

    @fasteners.locked(lock='_open_close_lock')
    def connect(self):
        """Opens the connection to any backend system."""
        if self._client is None:
            etcd_conf = {}
            for config_opts in self.ETCD_CONFIG_OPTIONS:
                key, value_type = config_opts
                if key in self._conf:
                    etcd_conf[key] = value_type(self._conf[key])

            self._client = etcd3gw.Etcd3Client(**etcd_conf)
            self._state = self.CONNECTED_STATE

            watch_url = self._create_path(self._root_path, self.JOB_PREFIX)
            self._thread_cancel = threading.Event()
            try:
                (self._watcher,
                 self._watcher_cancel) = self._client.watch_prefix(watch_url)
            except etcd3gw.exceptions.ConnectionFailedError:
                exc.raise_with_cause(exc.JobFailure,
                                     "Failed to connect to Etcd")
            self._watcher_thd = threading.Thread(target=self._watcher_thread)
            self._watcher_thd.start()

    def _watcher_thread(self):
        while not self._thread_cancel.is_set():
            for event in self._watcher:
                if "kv" not in event:
                    continue

                key_value = event["kv"]
                key = misc.binary_decode(key_value["key"])

                if key.endswith(self.DATA_POSTFIX):
                    key = key.rstrip(self.DATA_POSTFIX)
                    if event.get("type") == "DELETE":
                        self._remove_job_from_cache(key)
                    else:
                        data = key_value["value"]
                        self._process_incoming_job(key, data)

    @fasteners.locked(lock='_open_close_lock')
    def close(self):
        """Close the connection to any backend system."""
        if self._client is not None:
            if self._watcher_cancel is not None:
                self._watcher_cancel()
            if self._thread_cancel is not None:
                self._thread_cancel.set()
            if self._watcher_thd is not None:
                self._watcher_thd.join()
            del self._client
            self._client = None
            self._state = self.INIT_STATE



#    Copyright (C) 2012 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import taskflow.engines
from taskflow.patterns import linear_flow as lf
from taskflow.persistence.backends import impl_memory
from taskflow import task
from taskflow import test
from taskflow.utils import persistence_utils as p_utils


class ProgressTask(task.Task):
    def __init__(self, name, segments):
        super().__init__(name=name)
        self._segments = segments

    def execute(self):
        if self._segments <= 0:
            return
        for i in range(1, self._segments):
            progress = float(i) / self._segments
            self.update_progress(progress)


class ProgressTaskWithDetails(task.Task):
    def execute(self):
        details = {
            'progress': 0.5,
            'test': 'test data',
            'foo': 'bar',
        }
        self.notifier.notify(task.EVENT_UPDATE_PROGRESS, details)


class TestProgress(test.TestCase):
    def _make_engine(self, flow, flow_detail=None, backend=None):
        e = taskflow.engines.load(flow,
                                  flow_detail=flow_detail,
                                  backend=backend)
        e.compile()
        e.prepare()
        return e

    def tearDown(self):
        super().tearDown()
        with contextlib.closing(impl_memory.MemoryBackend({})) as be:
            with contextlib.closing(be.get_connection()) as conn:
                conn.clear_all()

    def test_sanity_progress(self):
        fired_events = []

        def notify_me(event_type, details):
            fired_events.append(details.pop('progress'))

        ev_count = 5
        t = ProgressTask("test", ev_count)
        t.notifier.register(task.EVENT_UPDATE_PROGRESS, notify_me)
        flo = lf.Flow("test")
        flo.add(t)
        e = self._make_engine(flo)
        e.run()
        self.assertEqual(ev_count + 1, len(fired_events))
        self.assertEqual(1.0, fired_events[-1])
        self.assertEqual(0.0, fired_events[0])

    def test_no_segments_progress(self):
        fired_events = []

        def notify_me(event_type, details):
            fired_events.append(details.pop('progress'))

        t = ProgressTask("test", 0)
        t.notifier.register(task.EVENT_UPDATE_PROGRESS, notify_me)
        flo = lf.Flow("test")
        flo.add(t)
        e = self._make_engine(flo)
        e.run()
        # 0.0 and 1.0 should be automatically fired
        self.assertEqual(2, len(fired_events))
        self.assertEqual(1.0, fired_events[-1])
        self.assertEqual(0.0, fired_events[0])

    def test_storage_progress(self):
        with contextlib.closing(impl_memory.MemoryBackend({})) as be:
            flo = lf.Flow("test")
            flo.add(ProgressTask("test", 3))
            b, fd = p_utils.temporary_flow_detail(be)
            e = self._make_engine(flo, flow_detail=fd, backend=be)
            e.run()
            end_progress = e.storage.get_task_progress("test")
            self.assertEqual(1.0, end_progress)
            task_uuid = e.storage.get_atom_uuid("test")
            td = fd.find(task_uuid)
            self.assertEqual(1.0, td.meta['progress'])
            self.assertFalse(td.meta['progress_details'])

    def test_storage_progress_detail(self):
        flo = ProgressTaskWithDetails("test")
        e = self._make_engine(flo)
        e.run()
        end_progress = e.storage.get_task_progress("test")
        self.assertEqual(1.0, end_progress)
        end_details = e.storage.get_task_progress_details("test")
        self.assertEqual(0.5, end_details.get('at_progress'))
        self.assertEqual({
            'test': 'test data',
            'foo': 'bar'
        }, end_details.get('details'))

    def test_dual_storage_progress(self):
        fired_events = []

        def notify_me(event_type, details):
            fired_events.append(details.pop('progress'))

        with contextlib.closing(impl_memory.MemoryBackend({})) as be:
            t = ProgressTask("test", 5)
            t.notifier.register(task.EVENT_UPDATE_PROGRESS, notify_me)
            flo = lf.Flow("test")
            flo.add(t)
            b, fd = p_utils.temporary_flow_detail(be)
            e = self._make_engine(flo, flow_detail=fd, backend=be)
            e.run()

            end_progress = e.storage.get_task_progress("test")
            self.assertEqual(1.0, end_progress)
            task_uuid = e.storage.get_atom_uuid("test")
            td = fd.find(task_uuid)
            self.assertEqual(1.0, td.meta['progress'])
            self.assertFalse(td.meta['progress_details'])
            self.assertEqual(6, len(fired_events))



#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import _thread
import collections
import multiprocessing
import threading

from taskflow.utils import misc


def is_alive(thread):
    """Helper to determine if a thread is alive (handles none safely)."""
    if not thread:
        return False
    return thread.is_alive()


def get_ident():
    """Return the 'thread identifier' of the current thread."""
    return _thread.get_ident()


def get_optimal_thread_count(default=2):
    """Try to guess optimal thread count for current system."""
    try:
        return multiprocessing.cpu_count() + 1
    except NotImplementedError:
        # NOTE(harlowja): apparently may raise so in this case we will
        # just setup two threads since it's hard to know what else we
        # should do in this situation.
        return default


def daemon_thread(target, *args, **kwargs):
    """Makes a daemon thread that calls the given target when started."""
    thread = threading.Thread(target=target, args=args, kwargs=kwargs)
    # NOTE(skudriashev): When the main thread is terminated unexpectedly
    # and thread is still alive - it will prevent main thread from exiting
    # unless the daemon property is set to True.
    thread.daemon = True
    return thread


# Container for thread creator + associated callbacks.
_ThreadBuilder = collections.namedtuple('_ThreadBuilder',
                                        ['thread_factory',
                                         'before_start', 'after_start',
                                         'before_join', 'after_join'])
_ThreadBuilder.fields = tuple([
    'thread_factory',
    'before_start',
    'after_start',
    'before_join',
    'after_join',
])


def no_op(*args, **kwargs):
    """Function that does nothing."""


class ThreadBundle:
    """A group/bundle of threads that start/stop together."""

    def __init__(self):
        self._threads = []
        self._lock = threading.Lock()

    def bind(self, thread_factory,
             before_start=None, after_start=None,
             before_join=None, after_join=None):
        """Adds a thread (to-be) into this bundle (with given callbacks).

        NOTE(harlowja): callbacks provided should not attempt to call
                        mutating methods (:meth:`.stop`, :meth:`.start`,
                        :meth:`.bind` ...) on this object as that will result
                        in dead-lock since the lock on this object is not
                        meant to be (and is not) reentrant...
        """
        if before_start is None:
            before_start = no_op
        if after_start is None:
            after_start = no_op
        if before_join is None:
            before_join = no_op
        if after_join is None:
            after_join = no_op
        builder = _ThreadBuilder(thread_factory,
                                 before_start, after_start,
                                 before_join, after_join)
        for attr_name in builder.fields:
            cb = getattr(builder, attr_name)
            if not callable(cb):
                raise ValueError("Provided callback for argument"
                                 " '%s' must be callable" % attr_name)
        with self._lock:
            self._threads.append([
                builder,
                # The built thread.
                None,
                # Whether the built thread was started (and should have
                # ran or still be running).
                False,
            ])

    def start(self):
        """Creates & starts all associated threads (that are not running)."""
        count = 0
        with self._lock:
            it = enumerate(self._threads)
            for i, (builder, thread, started) in it:
                if thread and started:
                    continue
                if not thread:
                    self._threads[i][1] = thread = builder.thread_factory()
                builder.before_start(thread)
                thread.start()
                count += 1
                try:
                    builder.after_start(thread)
                finally:
                    # Just incase the 'after_start' callback blows up make sure
                    # we always set this...
                    self._threads[i][2] = started = True
        return count

    def stop(self):
        """Stops & joins all associated threads (that have been started)."""
        count = 0
        with self._lock:
            it = misc.reverse_enumerate(self._threads)
            for i, (builder, thread, started) in it:
                if not thread or not started:
                    continue
                builder.before_join(thread)
                thread.join()
                count += 1
                try:
                    builder.after_join(thread)
                finally:
                    # Just incase the 'after_join' callback blows up make sure
                    # we always set/reset these...
                    self._threads[i][1] = thread = None
                    self._threads[i][2] = started = False
        return count

    def __len__(self):
        """Returns how many threads (to-be) are in this bundle."""
        return len(self._threads)



#    Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import taskflow.engines as engines
from taskflow.patterns import linear_flow
from taskflow import task as base
from taskflow import test


def double(x):
    return x * 2

square = lambda x: x * x


class MapFunctorTaskTest(test.TestCase):

    def setUp(self):
        super().setUp()

        self.flow_store = {
            'a': 1,
            'b': 2,
            'c': 3,
            'd': 4,
            'e': 5,
        }

    def test_double_array(self):
        expected = self.flow_store.copy()
        expected.update({
            'double_a': 2,
            'double_b': 4,
            'double_c': 6,
            'double_d': 8,
            'double_e': 10,
        })

        requires = self.flow_store.keys()
        provides = ["double_%s" % k for k in requires]

        flow = linear_flow.Flow("double array flow")
        flow.add(base.MapFunctorTask(double, requires=requires,
                                     provides=provides))

        result = engines.run(flow, store=self.flow_store)
        self.assertDictEqual(expected, result)

    def test_square_array(self):
        expected = self.flow_store.copy()
        expected.update({
            'square_a': 1,
            'square_b': 4,
            'square_c': 9,
            'square_d': 16,
            'square_e': 25,
        })

        requires = self.flow_store.keys()
        provides = ["square_%s" % k for k in requires]

        flow = linear_flow.Flow("square array flow")
        flow.add(base.MapFunctorTask(square, requires=requires,
                                     provides=provides))

        result = engines.run(flow, store=self.flow_store)
        self.assertDictEqual(expected, result)



#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import threading

import kombu
from kombu import exceptions as kombu_exceptions

from taskflow.engines.worker_based import dispatcher
from taskflow import logging

LOG = logging.getLogger(__name__)

# NOTE(skudriashev): A timeout of 1 is often used in environments where
# the socket can get "stuck", and is a best practice for Kombu consumers.
DRAIN_EVENTS_PERIOD = 1

# Helper objects returned when requested to get connection details, used
# instead of returning the raw results from the kombu connection objects
# themselves so that a person can not mutate those objects (which would be
# bad).
_ConnectionDetails = collections.namedtuple('_ConnectionDetails',
                                            ['uri', 'transport'])
_TransportDetails = collections.namedtuple('_TransportDetails',
                                           ['options', 'driver_type',
                                            'driver_name', 'driver_version'])


class Proxy:
    """A proxy processes messages from/to the named exchange.

    For **internal** usage only (not for public consumption).
    """

    DEFAULT_RETRY_OPTIONS = {
        # The number of seconds we start sleeping for.
        'interval_start': 1,
        # How many seconds added to the interval for each retry.
        'interval_step': 1,
        # Maximum number of seconds to sleep between each retry.
        'interval_max': 1,
        # Maximum number of times to retry.
        'max_retries': 3,
    }
    """Settings used (by default) to reconnect under transient failures.

    See: http://kombu.readthedocs.org/ (and connection ``ensure_options``) for
    what these values imply/mean...
    """

    # This is the only provided option that should be an int, the others
    # are allowed to be floats; used when we check that the user-provided
    # value is valid...
    _RETRY_INT_OPTS = frozenset(['max_retries'])

    def __init__(self, topic, exchange,
                 type_handlers=None, on_wait=None, url=None,
                 transport=None, transport_options=None,
                 retry_options=None):
        self._topic = topic
        self._exchange_name = exchange
        self._on_wait = on_wait
        self._running = threading.Event()
        self._dispatcher = dispatcher.TypeDispatcher(
            # NOTE(skudriashev): Process all incoming messages only if proxy is
            # running, otherwise requeue them.
            requeue_filters=[lambda data, message: not self.is_running],
            type_handlers=type_handlers)

        ensure_options = self.DEFAULT_RETRY_OPTIONS.copy()
        if retry_options is not None:
            # Override the defaults with any user provided values...
            for k in set(ensure_options.keys()):
                if k in retry_options:
                    # Ensure that the right type is passed in...
                    val = retry_options[k]
                    if k in self._RETRY_INT_OPTS:
                        tmp_val = int(val)
                    else:
                        tmp_val = float(val)
                    if tmp_val < 0:
                        raise ValueError("Expected value greater or equal to"
                                         " zero for 'retry_options' %s; got"
                                         " %s instead" % (k, val))
                    ensure_options[k] = tmp_val
        self._ensure_options = ensure_options

        self._drain_events_timeout = DRAIN_EVENTS_PERIOD
        if transport == 'memory' and transport_options:
            polling_interval = transport_options.get('polling_interval')
            if polling_interval is not None:
                self._drain_events_timeout = polling_interval

        # create connection
        self._conn = kombu.Connection(url, transport=transport,
                                      transport_options=transport_options)

        # create exchange
        self._exchange = kombu.Exchange(name=self._exchange_name,
                                        durable=False, auto_delete=True)

    @property
    def dispatcher(self):
        """Dispatcher internally used to dispatch message(s) that match."""
        return self._dispatcher

    @property
    def connection_details(self):
        """Details about the connection (read-only)."""
        # The kombu drivers seem to use 'N/A' when they don't have a version...
        driver_version = self._conn.transport.driver_version()
        if driver_version and driver_version.lower() == 'n/a':
            driver_version = None
        if self._conn.transport_options:
            transport_options = self._conn.transport_options.copy()
        else:
            transport_options = {}
        transport = _TransportDetails(
            options=transport_options,
            driver_type=self._conn.transport.driver_type,
            driver_name=self._conn.transport.driver_name,
            driver_version=driver_version)
        return _ConnectionDetails(
            uri=self._conn.as_uri(include_password=False),
            transport=transport)

    @property
    def is_running(self):
        """Return whether the proxy is running."""
        return self._running.is_set()

    def _make_queue(self, routing_key, exchange, channel=None):
        """Make a named queue for the given exchange."""
        queue_name = f"{self._exchange_name}_{routing_key}"
        return kombu.Queue(name=queue_name,
                           routing_key=routing_key, durable=False,
                           exchange=exchange, auto_delete=True,
                           channel=channel)

    def publish(self, msg, routing_key, reply_to=None, correlation_id=None):
        """Publish message to the named exchange with given routing key."""
        if isinstance(routing_key, str):
            routing_keys = [routing_key]
        else:
            routing_keys = routing_key

        # Filter out any empty keys...
        routing_keys = [r_k for r_k in routing_keys if r_k]
        if not routing_keys:
            LOG.warning("No routing key/s specified; unable to send '%s'"
                        " to any target queue on exchange '%s'", msg,
                        self._exchange_name)
            return

        def _publish(producer, routing_key):
            queue = self._make_queue(routing_key, self._exchange)
            producer.publish(body=msg.to_dict(),
                             routing_key=routing_key,
                             exchange=self._exchange,
                             declare=[queue],
                             type=msg.TYPE,
                             reply_to=reply_to,
                             correlation_id=correlation_id)

        def _publish_errback(exc, interval):
            LOG.exception('Publishing error: %s', exc)
            LOG.info('Retry triggering in %s seconds', interval)

        LOG.debug("Sending '%s' message using routing keys %s",
                  msg, routing_keys)
        with kombu.connections[self._conn].acquire(block=True) as conn:
            with conn.Producer() as producer:
                ensure_kwargs = self._ensure_options.copy()
                ensure_kwargs['errback'] = _publish_errback
                safe_publish = conn.ensure(producer, _publish, **ensure_kwargs)
                for routing_key in routing_keys:
                    safe_publish(producer, routing_key)

    def start(self):
        """Start proxy."""

        def _drain(conn, timeout):
            try:
                conn.drain_events(timeout=timeout)
            except kombu_exceptions.TimeoutError:
                pass

        def _drain_errback(exc, interval):
            LOG.exception('Draining error: %s', exc)
            LOG.info('Retry triggering in %s seconds', interval)

        LOG.info("Starting to consume from the '%s' exchange.",
                 self._exchange_name)
        with kombu.connections[self._conn].acquire(block=True) as conn:
            queue = self._make_queue(self._topic, self._exchange, channel=conn)
            callbacks = [self._dispatcher.on_message]
            with conn.Consumer(queues=queue, callbacks=callbacks) as consumer:
                ensure_kwargs = self._ensure_options.copy()
                ensure_kwargs['errback'] = _drain_errback
                safe_drain = conn.ensure(consumer, _drain, **ensure_kwargs)
                self._running.set()
                try:
                    while self._running.is_set():
                        safe_drain(conn, self._drain_events_timeout)
                        if self._on_wait is not None:
                            self._on_wait()
                finally:
                    self._running.clear()

    def wait(self):
        """Wait until proxy is started."""
        self._running.wait()

    def stop(self):
        """Stop proxy."""
        self._running.clear()



#    Copyright (C) 2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc

from taskflow.types import notifier
from taskflow.utils import misc


class Engine(metaclass=abc.ABCMeta):
    """Base for all engines implementations.

    :ivar Engine.notifier: A notification object that will dispatch
                    events that occur related to the flow the engine
                    contains.
    :ivar atom_notifier: A notification object that will dispatch events that
                         occur related to the atoms the engine contains.

    """

    def __init__(self, flow, flow_detail, backend, options):
        self._flow = flow
        self._flow_detail = flow_detail
        self._backend = backend
        self._options = misc.safe_copy_dict(options)
        self._notifier = notifier.Notifier()
        self._atom_notifier = notifier.Notifier()

    @property
    def notifier(self):
        """The flow notifier."""
        return self._notifier

    @property
    def atom_notifier(self):
        """The atom notifier."""
        return self._atom_notifier

    @property
    def options(self):
        """The options that were passed to this engine on construction."""
        return self._options

    @property
    @abc.abstractmethod
    def storage(self):
        """The storage unit for this engine."""

    @property
    @abc.abstractmethod
    def statistics(self):
        """A dictionary of runtime statistics this engine has gathered.

        This dictionary will be empty when the engine has never been
        ran. When it is running or has ran previously it should have (but
        may not) have useful and/or informational keys and values when
        running is underway and/or completed.

        .. warning:: The keys in this dictionary **should** be some what
                     stable (not changing), but there existence **may**
                     change between major releases as new statistics are
                     gathered or removed so before accessing keys ensure that
                     they actually exist and handle when they do not.
        """

    @abc.abstractmethod
    def compile(self):
        """Compiles the contained flow into a internal representation.

        This internal representation is what the engine will *actually* use to
        run. If this compilation can not be accomplished then an exception
        is expected to be thrown with a message indicating why the compilation
        could not be achieved.
        """

    @abc.abstractmethod
    def reset(self):
        """Reset back to the ``PENDING`` state.

        If a flow had previously ended up (from a prior engine
        :py:func:`.run`) in the ``FAILURE``, ``SUCCESS`` or ``REVERTED``
        states (or for some reason it ended up in an intermediary state) it
        can be desirable to make it possible to run it again. Calling this
        method enables that to occur (without causing a state transition
        failure, which would typically occur if :py:meth:`.run` is called
        directly without doing a reset).
        """

    @abc.abstractmethod
    def prepare(self):
        """Performs any pre-run, but post-compilation actions.

        NOTE(harlowja): During preparation it is currently assumed that the
        underlying storage will be initialized, the atoms will be reset and
        the engine will enter the ``PENDING`` state.
        """

    @abc.abstractmethod
    def validate(self):
        """Performs any pre-run, post-prepare validation actions.

        NOTE(harlowja): During validation all final dependencies
        will be verified and ensured. This will by default check that all
        atoms have satisfiable requirements (satisfied by some other
        provider).
        """

    @abc.abstractmethod
    def run(self):
        """Runs the flow in the engine to completion (or die trying)."""

    @abc.abstractmethod
    def suspend(self):
        """Attempts to suspend the engine.

        If the engine is currently running atoms then this will attempt to
        suspend future work from being started (currently active atoms can
        not currently be preempted) and move the engine into a suspend state
        which can then later be resumed from.
        """



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""initial_logbook_details_tables

Revision ID: 1cea328f0f65
Revises: None
Create Date: 2013-08-23 11:41:49.207087

"""

# revision identifiers, used by Alembic.
revision = '1cea328f0f65'
down_revision = None

import logging

from alembic import op
import sqlalchemy as sa

from taskflow.persistence.backends.sqlalchemy import tables

LOG = logging.getLogger(__name__)


def _get_indexes():
    # Ensure all uuids are indexed since they are what is typically looked
    # up and fetched, so attempt to ensure that that is done quickly.
    indexes = [
        {
            'index_name': 'logbook_uuid_idx',
            'table_name': 'logbooks',
            'columns': ['uuid'],
        },
        {
            'index_name': 'flowdetails_uuid_idx',
            'table_name': 'flowdetails',
            'columns': ['uuid'],
        },
        {
            'index_name': 'taskdetails_uuid_idx',
            'table_name': 'taskdetails',
            'columns': ['uuid'],
        },
    ]
    return indexes


def _get_foreign_keys():
    f_keys = [
        # Flow details uuid -> logbook parent uuid
        {
            'constraint_name': 'flowdetails_ibfk_1',
            'source_table': 'flowdetails',
            'referent_table': 'logbooks',
            'local_cols': ['parent_uuid'],
            'remote_cols': ['uuid'],
            'ondelete': 'CASCADE',
        },
        # Task details uuid -> flow details parent uuid
        {
            'constraint_name': 'taskdetails_ibfk_1',
            'source_table': 'taskdetails',
            'referent_table': 'flowdetails',
            'local_cols': ['parent_uuid'],
            'remote_cols': ['uuid'],
            'ondelete': 'CASCADE',
        },
    ]
    return f_keys


def upgrade():
    op.create_table('logbooks',
                    sa.Column('created_at', sa.DateTime),
                    sa.Column('updated_at', sa.DateTime),
                    sa.Column('meta', sa.Text(), nullable=True),
                    sa.Column('name', sa.String(length=tables.NAME_LENGTH),
                              nullable=True),
                    sa.Column('uuid', sa.String(length=tables.UUID_LENGTH),
                              primary_key=True, nullable=False),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8')
    op.create_table('flowdetails',
                    sa.Column('created_at', sa.DateTime),
                    sa.Column('updated_at', sa.DateTime),
                    sa.Column('parent_uuid',
                              sa.String(length=tables.UUID_LENGTH)),
                    sa.Column('meta', sa.Text(), nullable=True),
                    sa.Column('state', sa.String(length=tables.STATE_LENGTH),
                              nullable=True),
                    sa.Column('name', sa.String(length=tables.NAME_LENGTH),
                              nullable=True),
                    sa.Column('uuid', sa.String(length=tables.UUID_LENGTH),
                              primary_key=True, nullable=False),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8')
    op.create_table('taskdetails',
                    sa.Column('created_at', sa.DateTime),
                    sa.Column('updated_at', sa.DateTime),
                    sa.Column('parent_uuid',
                              sa.String(length=tables.UUID_LENGTH)),
                    sa.Column('meta', sa.Text(), nullable=True),
                    sa.Column('name', sa.String(length=tables.NAME_LENGTH),
                              nullable=True),
                    sa.Column('results', sa.Text(), nullable=True),
                    sa.Column('version',
                              sa.String(length=tables.VERSION_LENGTH),
                              nullable=True),
                    sa.Column('stacktrace', sa.Text(), nullable=True),
                    sa.Column('exception', sa.Text(), nullable=True),
                    sa.Column('state', sa.String(length=tables.STATE_LENGTH),
                              nullable=True),
                    sa.Column('uuid', sa.String(length=tables.UUID_LENGTH),
                              primary_key=True, nullable=False),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8')
    try:
        for fkey_descriptor in _get_foreign_keys():
            op.create_foreign_key(**fkey_descriptor)
    except NotImplementedError as e:
        LOG.warning("Foreign keys are not supported: %s", e)
    try:
        for index_descriptor in _get_indexes():
            op.create_index(**index_descriptor)
    except NotImplementedError as e:
        LOG.warning("Indexes are not supported: %s", e)


def downgrade():
    for table in ['logbooks', 'flowdetails', 'taskdetails']:
        op.drop_table(table)



#    Copyright (C) 2012-2013 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

logging.basicConfig(level=logging.ERROR)

top_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       os.pardir,
                                       os.pardir))
sys.path.insert(0, top_dir)

import taskflow.engines
from taskflow.patterns import graph_flow as gf
from taskflow import task

import example_utils as eu  # noqa


# In this example we demonstrate use of a target flow (a flow that only
# executes up to a specified target) to make an *oversimplified* pseudo
# build system. It pretends to compile all sources to object files and
# link them into an executable. It also can build docs, but this can be
# "switched off" via targeted flow special power -- ability to ignore
# all tasks not needed by its target.


class CompileTask(task.Task):
    """Pretends to take a source and make object file."""
    default_provides = 'object_filename'

    def execute(self, source_filename):
        object_filename = '%s.o' % os.path.splitext(source_filename)[0]
        print('Compiling %s into %s'
              % (source_filename, object_filename))
        return object_filename


class LinkTask(task.Task):
    """Pretends to link executable form several object files."""
    default_provides = 'executable'

    def __init__(self, executable_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._executable_path = executable_path

    def execute(self, **kwargs):
        object_filenames = list(kwargs.values())
        print('Linking executable %s from files %s'
              % (self._executable_path,
                 ', '.join(object_filenames)))
        return self._executable_path


class BuildDocsTask(task.Task):
    """Pretends to build docs from sources."""
    default_provides = 'docs'

    def execute(self, **kwargs):
        for source_filename in kwargs.values():
            print("Building docs for %s" % source_filename)
        return 'docs'


def make_flow_and_store(source_files, executable_only=False):
    flow = gf.TargetedFlow('build-flow')
    object_targets = []
    store = {}
    for source in source_files:
        source_stored = '%s-source' % source
        object_stored = '%s-object' % source
        store[source_stored] = source
        object_targets.append(object_stored)
        flow.add(CompileTask(name='compile-%s' % source,
                             rebind={'source_filename': source_stored},
                             provides=object_stored))
    flow.add(BuildDocsTask(requires=list(store.keys())))

    # Try this to see executable_only switch broken:
    object_targets.append('docs')
    link_task = LinkTask('build/executable', requires=object_targets)
    flow.add(link_task)
    if executable_only:
        flow.set_target(link_task)
    return flow, store


if __name__ == "__main__":
    SOURCE_FILES = ['first.c', 'second.cpp', 'main.cpp']
    eu.print_wrapped('Running all tasks:')
    flow, store = make_flow_and_store(SOURCE_FILES)
    taskflow.engines.run(flow, store=store)

    eu.print_wrapped('Building executable, no docs:')
    flow, store = make_flow_and_store(SOURCE_FILES, executable_only=True)
    taskflow.engines.run(flow, store=store)
